'''A mixin class for adding fields to catalog elements.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, List, Optional

from .catalog import CatalogNameError


class CatalogElementMixin():
    '''Contains properties required for all catalog elements'''
    _name: Optional[str] = None
    _tags: Optional[Dict[str, List[str]]] = None

    def __init__(self,
                 name: Optional[str] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 **unused_kwargs):
        super().__init__()
        tmp_name = name or self._name
        if tmp_name is not None:
            self._name = tmp_name.lower()
        if self._name is None:
            raise CatalogNameError(f'{self.__class__.__name__} needs to define _name.')

        if self._tags is None:
            tmp_tags = tags or {}
        else:
            tmp_tags = self._tags.copy()
            if tags is not None:
                tmp_tags.update(tags)
        self._tags = {}
        for key, val in tmp_tags.items():
            if isinstance(val, str):
                val = [val]
            if not isinstance(val, list):
                raise NotImplementedError
            self._tags[key.lower()] = [vv.lower() for vv in val]

    @property
    def name(self) -> str:
        '''The catalog name of the object'''
        assert self._name is not None
        return self._name

    @property
    def tags(self) -> Dict[str, List[str]]:
        '''The catalog tags of the object'''
        assert self._tags is not None
        return self._tags.copy()

    def _tag_to_bool(self, tagname: str, default: bool) -> bool:
        '''Interprets a list of tag values as a boolean.

        As of 01-03-2024, only looks at first value in list,
        and checks if it equals 'true' (case-insensitive).

        If the tag is not set, will return default.
        '''
        assert self._tags is not None
        if tagname in self._tags:
            return self._tags[tagname][0].lower() == 'true'
        return default



'''An instantiator for saving pipelines as JSON.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import json
from pathlib import Path
from typing import Any, Dict, List, Optional

from ..executor.executor import ExecutablePipeline
from ..instantiator.instantiator import Instantiator, InstantiatorError
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import StepDesignator
from ..templates.impl.pipeline_step import PipelineStep
from ..templates.impl.parallel_step import ParallelStep
from ..executor.executor_kind import ExecutorKind

from .json_loader import JsonLoader


class JsonInstantiator(Instantiator):
    '''An instantiator for saving pipelines as JSON.'''

    def instantiate(self, pipeline: BoundPipeline) -> ExecutablePipeline:
        raise NotImplementedError(
            'Currently JsonInstantiator is only used to save pipelines to disk.')

    def _serialize_pipeline_steps(
            self,
            pipeline: BoundPipeline,
            model_paths: Dict[StepDesignator, Path]) -> List[Dict[str, Any]]:
        '''Turn all steps in this pipeline into a JSON serializable object for saving.'''
        steps = []
        for step in pipeline.steps:
            steps.append(self._serialize_step(step, model_paths))
        return steps

    def _serialize_step(
            self,
            step: PipelineStep,
            model_paths: Dict[StepDesignator, Path]) -> Dict[str, Any]:
        '''Turn this step into a JSON serializable object for saving.'''

        if isinstance(step, ParallelStep):
            return self._serialize_parallel(step, model_paths)

        retval: Dict[str, Any] = {
            'type': 'step',
        }
        if step.filename is not None:
            retval['model_filename'] = str(step.filename)
        if step.opt_name is not None:
            retval['name'] = step.opt_name

        model_name = step.model_name
        assert model_name is not None, (
            f'BUG: step {step.opt_name} is not parallel but has no model.'
        )
        retval['algorithm_name'] = model_name

        if step.queried:
            retval['queried'] = True

        hyperparams = {
            param: step.model.param_to_json(param, py_value)
            for param, py_value in step.hyperparams().items()
        }
        if hyperparams:
            try:
                _ = json.dumps(hyperparams)
            except (TypeError, OverflowError) as err:
                raise InstantiatorError(
                    f'Something in hyperparams for pipeline step {step.filename}'
                    f'is not JSON serializable:\n{hyperparams}.\n'
                    f'Perhaps {step.model.__class__.__name__}._hyperparam_lookup'
                    ' needs to be extended?') from err
            retval["hyperparams"] = hyperparams
        return retval

    def _serialize_parallel(
            self,
            step: ParallelStep,
            model_paths: Dict[StepDesignator, Path]) -> Dict[str, Any]:
        '''Turn this parallel step, including its subpipelines,
        into a JSON serializable object for saving.'''
        retval = {
            'type': 'parallel',
            'subpipelines': {}
        }
        for sub_name, sub_pipe in step.subpipelines.items():
            assert isinstance(sub_pipe, BoundPipeline), (
                f'BUG: subpipeline {sub_name} recieved by JsonInstantiator is not '
                f'a BoundPipeline, instead found {type(sub_pipe)}. ')
            assert isinstance(retval['subpipelines'], dict)
            retval['subpipelines'][sub_name] = self._serialize_pipeline_steps(
                pipeline=sub_pipe,
                model_paths=model_paths)
        return retval

    def save(self, pipeline: BoundPipeline, model_paths: Dict[StepDesignator, Path]) -> Path:
        '''Save pipeline as a JSON file using self._saver.

        Returns Path to the saved file.
        '''
        if self._saver is None:
            raise InstantiatorError('You must define a Saver to save from an Instantiator.')

        image = {
            'version': self._saver.current_version,
            'pipeline_designator': pipeline.designator,
            'pipeline_template_name': pipeline.name,
            'output_dir': str(self._saver.output_path),
            'steps': self._serialize_pipeline_steps(pipeline=pipeline, model_paths=model_paths)
        }

        body = json.dumps(image, indent=4).encode()

        return self._saver.save_pipeline(
            body,
            des=pipeline.designator,
            kind=ExecutorKind('json'))

    def load(self, pipeline_file: Path, load_models: bool = False,
             model_folder: Optional[Path] = None) -> BoundPipeline:
        '''Load a BoundPipeline from disk.'''
        assert self._algorithm_catalog is not None, (
            'BUG: loading a pipeline requires an algorithm catalog.'
        )
        assert self._saver is not None, (
            'BUG: loading a pipeline requires a saver.'
        )
        return JsonLoader(saver_version=self._saver.current_version,
                          algorithm_catalog=self._algorithm_catalog,
                          load_models=load_models,
                          model_folder=model_folder,
                          pipeline_file=pipeline_file).pipeline



'''A fake algorithm used for tests.'''

from typing import Optional

from .algorithm import Algorithm
from .fittable_algorithm_instance import FittableAlgorithmInstance

from ...wrangler.dataset import Dataset


# TODO(Merritt/Piggy): possibly unify this with more uses of FakeAlgorithm in tests
class FakeAlgorithm(Algorithm):
    '''A fake algorithm used for tests.'''
    _name = 'fake_algorithm'

    def instantiate(self, **hyperparams) -> 'FakeInstance':
        return FakeInstance(parent=self, **hyperparams)


FAKE_SERIALIZED = b'fake serialized algorithm'


class FakeInstance(FittableAlgorithmInstance):
    '''An instance of FakeAlgorithm for tests.'''
    evidence_of_deserialize = None

    def __init__(self, parent, **kwargs):
        super().__init__(parent=parent, **kwargs)

    def deserialize(self, serialized_model: bytes) -> 'FakeInstance':
        self.evidence_of_deserialize = serialized_model
        return self

    def serialize(self) -> bytes:
        return FAKE_SERIALIZED

    def fit(self, dataset: Optional[Dataset]) -> None:
        _ = dataset
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        return dataset



'''Catalog for objects communicate with neighbors in distributed algorithms.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, List, Optional, Type, Union

from ...catalog.memory_catalog import MemoryCatalog

from .communicator import Communicator


class CommunicatorCatalog(MemoryCatalog[Type[Communicator]], metaclass=abc.ABCMeta):
    '''Base class for Communicator catalogs'''

    def register(self, obj: Type[Communicator], name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        return super().register(obj, name, tags)


class CommunicatorCatalogStub(CommunicatorCatalog):
    '''stub'''



'''Wrapper for sklearn.impute.SimpleImputer'''
import pickle
from typing import Dict, Any, List, Optional

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from sklearn import impute  # type: ignore[import]
import pandas as pd
import numpy as np

from ....algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ....algorithms.impl.algorithm_instance import AlgorithmInstance, DatasetError
from ....algorithms.impl.fittable_algorithm_instance import (FittableAlgorithmInstance,
                                                             UntrainedError)
from ....catalog.catalog import upcast
from ....problem_def.task import DataType
from ....wrangler.dataset import Dataset
from ....wrangler.constants import ColumnType


class SimpleImputerModel(Algorithm):
    '''Wrapper for sklearn.impute.SimpleImputer

    Different imputers are used for different column types (numeric, integral, categorical).
    Each hyperparam is a dict of subhyperparams, which map 1:1 onto the sklearn imputers.

    ===========
    hyperparams
    ===========

    :numeric: Dict[str, Any]

        Hyperparams for the sklearn.impute.SimpleImputer applied to continuous or
        real-valued numeric columns.

        Default\\:

        .. code-block:: Python

            {
                'missing_values': np.nan,
                'strategy': 'mean',
                'fill_value': None,
                'copy': True,
                'add_indicator': False,
                'keep_empty_features': False
            }

    :integral: Dict[str, Any]

        Hyperparams for the sklearn.impute.SimpleImputer applied to discrete or integer columns.

        Default\\:

        .. code-block:: Python

            {
                'missing_values': np.nan,
                'strategy': 'median',
                'fill_value': None,
                'copy': True,
                'add_indicator': False,
                'keep_empty_features': False
            }

    :categorical: Dict[str, Any]

        Hyperparams for the sklearn.impute.SimpleImputer applied to categorical columns.

        Default\\:

        .. code-block:: Python

            {
                'missing_values': None,
                'strategy': 'most_frequent',
                'fill_value': None,
                'copy': True,
                'add_indicator': False,
                'keep_empty_features': False
            }

    ==============
    subhyperparams
    ==============

    :missing_values: int, float, str, np.nan, None or pandas.NA, default=np.nan

        The placeholder for the missing values.
        All occurrences of missing_values will be imputed. For pandas dataframes
        with nullable integer dtypes with missing values, missing_values can
        be set to either np.nan or pd.NA.

    :strategy: str, default='mean'

        The imputation strategy.

        If “mean”, then replace missing values using the mean along each
        column. Can only be used with numeric data.

        If “median”, then replace missing values using the median along each
        column. Can only be used with numeric data.

        If “most_frequent”, then replace missing using the most frequent
        value along each column. Can be used with strings or numeric data.
        If there is more than one such value, only the smallest is returned.

        If “constant”, then replace missing values with fill_value.
        Can be used with strings or numeric data.

    :fill_value: str or numerical value, default=None

        When strategy == “constant”, fill_value is used to replace all
        occurrences of missing_values. For string or object data types,
        fill_value must be a string. If None, fill_value will be 0 when imputing
        numerical data and “missing_value” for strings or object data types.

    :copy: bool, default=True

        If True, a copy of X will be created. If False, imputation will be
        done in-place whenever possible.

    :add_indicator: bool, default=False

        If True, a MissingIndicator transform will stack onto output of the
        imputer's transform. This allows a predictive estimator to account for
        missingness despite imputation. If a feature has no missing values at
        fit/train time, the feature won't appear on the missing indicator even
        if there are missing values at transform/test time.

    :keep_empty_features: bool, default=False

        If True, features that consist exclusively of missing values when fit is
        called are returned in results when transform is called. The imputed
        value is always 0 except when strategy="constant" in which case
        fill_value will be used instead.
    '''
    _name = 'sklearn.impute.SimpleImputer'
    _tags = {
        'source': ['sklearn'],
        'preprocessor': ['true'],
        'data_type': [DataType.TABULAR.name]
    }
    _default_hyperparams = {
        ColumnType.NUMERIC.value: {
            'missing_values': np.nan,
            'strategy': 'mean',
            'fill_value': None,
            'copy': True,
            'add_indicator': False,
            'keep_empty_features': False
        },
        ColumnType.INTEGRAL.value: {
            'missing_values': np.nan,
            'strategy': 'median',
            'fill_value': None,
            'copy': True,
            'add_indicator': False,
            'keep_empty_features': False
        },
        ColumnType.CATEGORICAL.value: {
            'missing_values': None,
            'strategy': 'most_frequent',
            'fill_value': None,
            'copy': True,
            'add_indicator': False,
            'keep_empty_features': False
        }
    }

    def instantiate(self, **hyperparams) -> AlgorithmInstance:
        return SimpleImputerModelInstance(parent=self, **hyperparams)


class SimpleImputerModelInstance(FittableAlgorithmInstance):
    '''Wrapper for sklearn.impute.SimpleImputer'''
    _imputers: Dict[ColumnType, impute.SimpleImputer]

    def __init__(self,
                 parent: SimpleImputerModel,
                 **hyperparams: Dict[str, Any]):
        super().__init__(parent=parent)
        self._imputers = self._make_imputers(**self.algorithm.hyperparams(**hyperparams))

    @classmethod
    def _make_imputers(cls, **hyperparams: Dict[str, Any]
                       ) -> Dict[ColumnType, impute.SimpleImputer]:
        retval: Dict[ColumnType, impute.SimpleImputer] = {}
        for col_type, params in hyperparams.items():
            imputer = impute.SimpleImputer(**params)
            imputer.set_output(transform='pandas')
            retval[ColumnType.from_str(col_type)] = imputer
        return retval

    @classmethod
    def _split_df_by_col_type(cls,
                              data_df: pd.DataFrame
                              ) -> Dict[ColumnType, pd.DataFrame]:
        # Sort columns by name to ensure stable order.
        cols = data_df.columns.to_list()
        data_df = data_df[sorted(cols)]

        numeric_df = data_df.select_dtypes(include=['float'])
        integral_df = data_df.select_dtypes(include=['int'])
        categorical_df = data_df.select_dtypes(include=['object', 'bool'])
        other_df = data_df.select_dtypes(exclude=['float', 'object', 'bool', 'int'])

        return {
            ColumnType.NUMERIC: numeric_df,
            ColumnType.INTEGRAL: integral_df,
            ColumnType.CATEGORICAL: categorical_df,
            ColumnType.OTHER: other_df
        }

    def _fit_imputers(self, data_split: Dict[ColumnType, pd.DataFrame]):
        # This may throw an error if re-fitting the same model on data with different column types
        for col_type in [ColumnType.NUMERIC, ColumnType.INTEGRAL, ColumnType.CATEGORICAL]:
            if not data_split[col_type].empty:
                self._imputers[col_type].fit(X=data_split[col_type])
            elif col_type in self._imputers:
                del self._imputers[col_type]

        # ignore ColumnType.OTHER
        self._trained = True

    def fit(self, dataset: Optional[Dataset]) -> None:
        '''Imputers can be fit, but they don't need to.'''
        if dataset is None:
            raise DatasetError(f'attempt to fit {self.catalog_name} with no data')
        data_df = dataset.dataframe
        data_split = self._split_df_by_col_type(data_df=data_df)
        self._fit_imputers(data_split=data_split)

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None

        data_df = dataset.dataframe
        data_split = self._split_df_by_col_type(data_df=data_df)

        if not self._trained:
            self._fit_imputers(data_split=data_split)

        output_list: List[pd.DataFrame] = [data_split[ColumnType.OTHER]]
        # do not transform cols with type OTHER
        for ctype, imp in self._imputers.items():
            if not data_split[ctype].empty:
                output_list.append(imp.transform(X=data_split[ctype]))

        output_df: pd.DataFrame = pd.concat(output_list, axis=1)
        retval = dataset.output()
        retval.dataframe = output_df
        return retval

    def deserialize(self, serialized_model: bytes) -> 'SimpleImputerModelInstance':
        self._imputers = pickle.loads(serialized_model)
        self._trained = True
        return self

    def serialize(self) -> bytes:
        '''Return a serialized version of a trained model.'''
        if not self._trained:
            raise UntrainedError(
                f'attempt to serialize model before fit for {self.algorithm.name}')
        return pickle.dumps(self._imputers, pickle.HIGHEST_PROTOCOL)


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = SimpleImputerModel(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Matcher functions'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Any, Dict, List

from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..templates.impl.pipeline_step import PipelineStep
from ..wrangler.constants import Matcher


class MatcherClass(metaclass=abc.ABCMeta):
    '''Matches a step or pipeline.

    args defines what this class matches
    matches(step, pipeline) returns True iff that step and pipeline match with args
    '''
    args: Any

    def __init__(self, args):
        self._args = args

    @abc.abstractmethod
    def matches(self, step: PipelineStep, pipeline: BoundPipeline) -> bool:
        '''Returns True iff step and pipeline match with this class's args'''


class AlgorithmMatcher(MatcherClass):
    '''Match by algorithm name.

    args is the algorithm name to match. This is the name of the algorithm
    associated with the step.
    '''
    _args: str

    def matches(self, step: PipelineStep, pipeline: BoundPipeline) -> bool:
        return step.has_model() and step.model_name == self._args


class DesignatorMatcher(MatcherClass):
    '''Match by designator.

    args is the designator to match.'''
    _args: Designator

    def matches(self, step: PipelineStep, pipeline: BoundPipeline) -> bool:
        designator = Designator(f'{pipeline.designator}')
        return designator == self._args


class NameMatcher(MatcherClass):
    '''Match by step name.

    args is the step name to match.'''
    _args: str

    def matches(self, step: PipelineStep, pipeline: BoundPipeline) -> bool:
        return step.pipeline_designator_component == self._args


class TagsMatcher(MatcherClass):
    '''Match by tags.

    args is the tag(s) to match

    matches() returns True iff at least one tag-value in args matches one of the
    corresponding tag-values on the algorithm in the pipeline step.
    '''

    _args: Dict[str, str]

    def matches(self, step: PipelineStep, pipeline: BoundPipeline) -> bool:
        tags = self._args
        if step.tags is None:
            # If the step has no model, it can not have tags and so can not match.
            return False
        submatches: List[bool] = []
        for tag_type, tag_value in tags.items():
            submatches.append(tag_type in step.tags and tag_value in step.tags[tag_type])
        return all(submatches)


_MATCHER_TABLE = {
    Matcher.ALGORITHM: AlgorithmMatcher,
    Matcher.DESIGNATOR: DesignatorMatcher,
    Matcher.NAME: NameMatcher,
    Matcher.TAGS: TagsMatcher,
}


class MatcherFactory:
    '''Select a matcher by Matcher'''
    def make(self, matcher: Matcher, args: Any) -> MatcherClass:
        '''Create a matcher class instance for the given matcher and args.'''
        # All of the classes in _MATCHER_TABLE define matches().
        return _MATCHER_TABLE[matcher](args)  # type: ignore[abstract]



'''
This module defines the WranglerResult class, which represents
the results of the Wrangler.fit_predict_rank function.

The WranglerResult class contains information about the
split dataset, pipeline results, rankings, and executable pipelines.
'''

from typing import Dict, Optional

from ..generator.designator import Designator
from ..instantiator.executable_pipeline import ExecutablePipeline, PipelineResults
from ..ranker.ranker import Rankings
from ..splitters.impl.splitter import SplitDataset


class WranglerResult():
    '''The results of Wrangler.fit_predict_rank()'''

    _split_dataset: SplitDataset
    _train_results: PipelineResults
    _test_results: Optional[PipelineResults]
    _rankings: Rankings

    def __init__(self,
                 split_dataset: SplitDataset,
                 train_results: PipelineResults,
                 test_results: Optional[PipelineResults],
                 rankings: Rankings):
        '''
        Initialize a WranglerResult object.

        Args:
        :split_dataset: The ground truth and folds used to rank the pipelines.
        :train_results: Predictions on train data, acquired using cross-validation.
        :test_results: Predictions on test data, if supplied in the problem definition.
        :rankings: The rankings of all bound pipelines.
        '''
        self._split_dataset = split_dataset
        self._train_results = train_results
        self._test_results = test_results
        self._rankings = rankings

    @property
    def split_dataset(self) -> SplitDataset:
        '''The ground truth and folds used to rank the pipelines.'''
        return self._split_dataset

    @property
    def train_results(self) -> PipelineResults:
        '''Predictions on train data, acquired using cross-validation.'''
        return self._train_results

    @property
    def test_results(self) -> Optional[PipelineResults]:
        '''Predictions on test data, if supplied in the problem definition.'''
        return self._test_results

    @property
    def rankings(self) -> Rankings:
        '''The rankings of all bound pipelines.'''
        return self._rankings

    @property
    def executable_pipelines(self) -> Dict[Designator, ExecutablePipeline]:
        '''Get all the executable pipelines.'''
        return self._train_results.executable_pipelines



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

##########################################
# Functions for rank aggregation
##########################################

import numpy as np
from typing import Optional, Tuple
from sklearn.preprocessing import MinMaxScaler  # type: ignore[import]
import cvxpy as cp
from itertools import combinations, permutations
from sklearn.neighbors import NearestNeighbors  # type: ignore[import]
import networkx as nx  # type: ignore[import]
from scipy.spatial.distance import pdist, squareform  # type: ignore[import]
from sklearn.cluster import AgglomerativeClustering  # type: ignore[import]

from .distributions import mallows_kendall as mk

##########################################
# Trimmed Rank Aggregators
##########################################


def trimmed_partial_borda(ranks: np.ndarray,
                          weights: Optional[np.ndarray] = None,
                          top_k: Optional[int] = None,
                          top_kr: Optional[int] = None,
                          aggregation_type='kemeny',
                          metric: str = 'influence',
                          n_neighbors: int = 6) -> Tuple[float, np.ndarray]:
    """Computes the trimmed borda rank

    Parameters
    ----------
    ranks: [# permutations, # items]
        Array of ranks

    weights: [# permutations,]
        Weights of each permutation. By default, weights=None.

    top_k: int
        How many items to consider for partial rank aggregation.
        By default top_k=None.

    top_kr: int
        How many permutations to use for rank aggregation.
        By default top_kr=None. If top is None, then use
        agglomerative clustering.

    aggregation_type: int
        Type of aggregation method to use while computing influence.
        We recommend 'borda' in large problems and kemeny in smaller
        problems.

    metric: str
        Metric of rank reliablity. By default metric='influence'.

    n_neighbors: int
        Number of neighbours to use for proximity based reliability
    """
    reliability = _get_reliability(ranks=ranks,
                                   metric=metric,
                                   aggregation_type=aggregation_type,
                                   top_k=top_k,
                                   n_neighbors=n_neighbors)

    if top_kr is None:
        trimmed_ranks = _get_trimmed_ranks_clustering(ranks, reliability)
    else:
        trimmed_ranks = ranks[np.argsort(-1 * reliability)[:top_kr], :]

    if weights is not None:
        trimmed_weights = weights[np.argsort(-1 * reliability)[:top_kr], :]
    else:
        trimmed_weights = None

    assert top_k is not None, 'BUG: top_k in trimmed_partial_borda came out as None'
    return partial_borda(ranks=trimmed_ranks,
                         weights=trimmed_weights,
                         top_k=top_k)


def trimmed_borda(ranks: np.ndarray,
                  weights: Optional[np.ndarray] = None,
                  top_k: Optional[int] = None,
                  top_kr: Optional[int] = None,
                  aggregation_type='kemeny',
                  metric: str = 'influence',
                  n_neighbors: int = 6) -> Tuple[float, np.ndarray]:
    """Computes the trimmed borda rank

    Parameters
    ----------
    ranks: [# permutations, # items]
        Array of ranks

    weights: [# permutations,]
        Weights of each permutation. By default, weights=None.

    top_k: int
        How many items to consider for partial rank aggregation.
        By default top_k=None.

    top_kr: int
        How many permutations to use for rank aggregation.
        By default top_kr=None. If top is None, then use
        agglomerative clustering.

    aggregation_type: int
        Type of aggregation method to use while computing influence.
        We recommend 'borda' in large problems and kemeny in smaller
        problems.

    metric: str
        Metric of rank reliablity. By default metric='influence'.

    n_neighbors: int
        Number of neighbours to use for proximity based reliability
    """
    reliability = _get_reliability(ranks=ranks,
                                   metric=metric,
                                   aggregation_type=aggregation_type,
                                   top_k=top_k,
                                   n_neighbors=n_neighbors)

    if top_kr is None:
        trimmed_ranks = _get_trimmed_ranks_clustering(ranks, reliability)
    else:
        trimmed_ranks = ranks[np.argsort(-1 * reliability)[:top_kr], :]

    if weights is not None:
        trimmed_weights = weights[np.argsort(-1 * reliability)[:top_kr], :]
    else:
        trimmed_weights = None

    return borda(ranks=trimmed_ranks, weights=trimmed_weights)


def trimmed_kemeny(ranks: np.ndarray,
                   weights: Optional[np.ndarray] = None,
                   top_k: Optional[int] = None,
                   top_kr: Optional[int] = None,
                   aggregation_type='kemeny',
                   metric: str = 'influence',
                   n_neighbors: int = 6,
                   verbose: bool = True) -> Tuple[float, np.ndarray]:
    """Computes the trimmed kemeny rank

    Parameters
    ----------
    ranks: [# permutations, # items]
        Array of ranks

    weights: [# permutations,]
        Weights of each permutation. By default, weights=None.

    top_k: int
        How many items to consider for partial rank aggregation.
        By default top_k=None.

    top_kr: int
        How many permutations to use for rank aggregation.
        By default top_kr=None. If top is None, then use
        agglomerative clustering.

    aggregation_type: int
        Type of aggregation method to use while computing influence.
        We recommend 'borda' in large problems and kemeny in smaller
        problems.

    metric: str
        Metric of rank reliablity. By default metric='influence'.

    n_neighbors: int
        Number of neighbours to use for proximity based reliability

    verbose: bool
        Controls verbosity
    """
    reliability = _get_reliability(ranks=ranks,
                                   metric=metric,
                                   aggregation_type=aggregation_type,
                                   top_k=top_k,
                                   n_neighbors=n_neighbors)

    if top_kr is None:
        trimmed_ranks = _get_trimmed_ranks_clustering(ranks, reliability)
    else:
        trimmed_ranks = ranks[np.argsort(-1 * reliability)[:top_kr], :]

    if weights is not None:
        trimmed_weights = weights[np.argsort(-1 * reliability)[:top_kr], :]
    else:
        trimmed_weights = None

    return kemeny(ranks=trimmed_ranks,
                  weights=trimmed_weights,
                  verbose=verbose)


##########################################
# Rank Aggregators
##########################################


def partial_borda(ranks: np.ndarray,
                  weights: Optional[np.ndarray] = None,
                  top_k: int = 5) -> Tuple[float, np.ndarray]:
    # Top-k Borda Rank Aggregation
    # NOTE: weights is only for compatibility, currently not using weights

    ranks = ranks.astype(float)
    ranks = np.nan_to_num(x=ranks,
                          nan=ranks.shape[1] + 1)  # If ranks already have nans
    # Mask higher ranks
    x, y = np.where((ranks > (top_k - 1)))
    for x_i, y_i in zip(x, y):
        ranks[x_i, y_i] = np.nan
    aggregated_rank = np.nan_to_num(x=mk.borda_partial(ranks, w=1, k=top_k),
                                    nan=ranks.shape[1] - 1).astype(int)
    objective = np.mean([mk.distance(r, aggregated_rank) for r in ranks])

    return objective, aggregated_rank


def borda(ranks: np.ndarray,
          weights: Optional[np.ndarray] = None) -> Tuple[float, np.ndarray]:
    if weights is None:
        aggregated_rank = mk.median(ranks)
    else:
        # aggregated_rank = mk.weighted_median(ranks)
        raise NotImplementedError(
            'borda can not do weights because mk does not implment it')
    objective = np.mean([mk.distance(r, aggregated_rank) for r in ranks])
    return objective, aggregated_rank


def kemeny(ranks: np.ndarray,
           weights: Optional[np.ndarray] = None,
           verbose: bool = True) -> Tuple[float, np.ndarray]:
    """Kemeny-Young optimal rank aggregation [1]

    We include the ability to incorporate weights of metrics/permutations.

    Parameters
    ----------
    ranks:
        Permutations/Ranks
    weights:
        Weight of each rank/permutation.
    verbose:
        Controls verbosity

    References
    ----------
    [1] Conitzer, V., Davenport, A., & Kalagnanam, J. (2006, July). Improved bounds for computing Kemeny rankings. In AAAI (Vol. 6, pp. 620-626).
        https://www.aaai.org/Papers/AAAI/2006/AAAI06-099.pdf
    [2] http://vene.ro/blog/kemeny-young-optimal-rank-aggregation-in-python.html#note4\
    """

    _, n_models = ranks.shape

    # Minimize C.T * X
    edge_weights = build_graph(ranks, weights)
    C = -1 * edge_weights.ravel().reshape((-1, 1))

    # Defining variables
    X = cp.Variable((n_models**2, 1), boolean=True)

    # Defining the objective function
    objective = cp.Maximize(C.T @ X)

    # Defining the constraints
    def idx(i, j):
        return n_models * i + j

    # Constraints for every pair
    pairwise_constraints = np.zeros(
        ((n_models * (n_models - 1)) // 2, n_models**2))
    for row, (i, j) in zip(pairwise_constraints,
                           combinations(range(n_models), 2)):
        row[[idx(i, j), idx(j, i)]] = 1

    # and for every cycle of length 3
    triangle_constraints = np.zeros(
        ((n_models * (n_models - 1) * (n_models - 2)), n_models**2))
    for row, (i, j, k) in zip(triangle_constraints,
                              permutations(range(n_models), 3)):
        row[[idx(i, j), idx(j, k), idx(k, i)]] = 1

    constraints = []
    constraints += [
        pairwise_constraints @ X == np.ones((pairwise_constraints.shape[0], 1))
    ]
    constraints += [
        triangle_constraints @ X >= np.ones((triangle_constraints.shape[0], 1))
    ]

    # Solving the problem
    problem = cp.Problem(objective, constraints)

    if verbose:
        print("Is DCP:", problem.is_dcp())
    problem.solve(verbose=verbose, warm_start=True)

    assert X.value is not None, 'BUG: X.value is None in kemeny'
    aggregated_rank: np.ndarray = X.value.reshape((n_models, n_models)).sum(axis=1)

    objective_value = np.mean([mk.distance(r, aggregated_rank) for r in ranks])

    return objective_value, aggregated_rank


##########################################
# Helper functions
##########################################


def _get_reliability(ranks,
                     metric='influence',
                     aggregation_type='borda',
                     top_k=None,
                     n_neighbors=6):
    if metric == 'influence':
        reliability = influence(ranks,
                                aggregation_type=aggregation_type,
                                top_k=top_k)
    elif metric == 'proximity':
        reliability = proximity(ranks, n_neighbors=n_neighbors, top_k=top_k)
    elif metric == 'pagerank':
        reliability = pagerank(ranks, top_k=top_k)
    elif metric == 'averagedistance':
        reliability = averagedistance(ranks, top_k=top_k)
    return reliability


def _get_trimmed_ranks_clustering(ranks, reliability):
    clustering = AgglomerativeClustering(n_clusters=2,
                                         linkage='single').fit_predict(
                                             reliability.reshape((-1, 1)))

    # cluster_ids, counts = np.unique(clustering, return_counts=True)
    # largest_cluster_idx = cluster_ids[np.argmax(counts)]  # Largest cluster

    most_reliable_cluster_idx = np.argmax([
        np.sum(reliability[np.where(clustering == 0)[0]]),
        np.sum(reliability[np.where(clustering == 1)[0]])
    ])
    # np.sum(reliability[np.where(clustering == 2)[0]])])

    # trimmed_ranks = ranks[np.where(clustering == largest_cluster_idx)[0], :]
    trimmed_ranks = ranks[np.where(clustering == most_reliable_cluster_idx)
                          [0], :]  # <--- NOTE: We used this
    # trimmed_ranks = ranks[reliability > 0, :]

    return trimmed_ranks


def compute_weights(ranks: np.ndarray,
                    true_rank: Optional[np.ndarray] = None) -> np.ndarray:
    """Computes the weight of a data point based on its distance from the true permutation.
    """
    n_metrics, n_models = ranks.shape
    if true_rank is None:
        true_rank = np.arange(n_models)
    distance_from_true_rank = np.array(
        [mk.distance(perm, true_rank) for perm in ranks])
    scaler = MinMaxScaler(feature_range=(0, 9))
    weights = scaler.fit_transform(distance_from_true_rank.reshape((-1, 1)))
    weights = 1 / (1 + weights)
    return weights.reshape((-1, 1))


def build_graph(ranks: np.ndarray,
                metric_weights: Optional[np.ndarray] = None) -> np.ndarray:
    n_metrics, n_models = ranks.shape
    if metric_weights is None:
        metric_weights = np.ones((n_metrics, 1))
    else:
        metric_weights = metric_weights.reshape((-1, 1))
    edge_weights = np.zeros((n_models, n_models))

    for i, j in combinations(range(n_models), 2):
        preference = ranks[:, i] - ranks[:, j]
        h_ij = (metric_weights.T @ (preference < 0).astype(int).reshape(
            (-1, 1))).squeeze()  # prefers i to j
        h_ji = (metric_weights.T @ (preference > 0).astype(int).reshape(
            (-1, 1))).squeeze()  # prefers j to i
        if h_ij > h_ji:
            edge_weights[i, j] = h_ij - h_ji
        elif h_ij < h_ji:
            edge_weights[j, i] = h_ji - h_ij
    return edge_weights


##########################################
# Functions to compute influence
##########################################


def objective(ranks, aggregation_type='kemeny', top_k=None):
    if aggregation_type == 'borda':
        _, sigma_star = borda(ranks=ranks, weights=None)
    elif aggregation_type == 'kemeny':
        _, sigma_star = kemeny(ranks=ranks, weights=None, verbose=False)
    elif aggregation_type == 'partial_borda':
        _, sigma_star = partial_borda(ranks=ranks, weights=None, top_k=top_k)
    return np.mean([mk.distance(r, sigma_star) for r in ranks])


def influence(ranks, aggregation_type='kemeny', top_k=None) -> np.ndarray:
    """Computes the reciprocal influence of each permutation/rank on the objective. Ranks with
    higher influence (and lower reciprocal influence) are more outlying.
    """
    N, n = ranks.shape
    objective_values = []

    if (aggregation_type == 'partial_borda') and (top_k is None):
        raise ValueError("top_k must be specified!")

    objective_all = objective(
        ranks, aggregation_type=aggregation_type,
        top_k=top_k)  # Objective when using all the permutations

    for i in combinations(np.arange(N), N - 1):
        objective_values.append(
            objective(ranks[i, :],
                      aggregation_type=aggregation_type,
                      top_k=top_k))

    # If removing a permutation results in a higher decrease in the objective
    # then it is more likely to be outlying
    influence = objective_all - np.array(
        objective_values[::-1])  # Reverse the list
    reliability = -influence

    # influence --
    # +ve -- metric good
    # -ve influence is bad
    # low positive influence or high positive influence?

    return reliability


def proximity(ranks, n_neighbors: int = 6, top_k=None) -> np.ndarray:
    """Computes the proximity of each rank to its nearest neighbours. Ranks with higher proximity are more central.
    """
    if top_k is not None:
        ranks = ranks.astype(float)
        x, y = np.where((ranks > (top_k - 1)))
        for x_i, y_i in zip(x, y):
            ranks[x_i, y_i] = np.nan

    neigh = NearestNeighbors(n_neighbors=n_neighbors,
                             algorithm='ball_tree',
                             metric=mk.distance)
    neigh.fit(ranks)

    proximity = 1 / neigh.kneighbors(ranks)[0].mean(axis=1)

    return proximity


def pagerank(ranks, top_k=None) -> np.ndarray:
    """Computes the pagerank of each rank. Higher pagerank implies that a rank is more authoritative.
    """
    if top_k is not None:
        ranks = ranks.astype(float)
        x, y = np.where((ranks > (top_k - 1)))
        for x_i, y_i in zip(x, y):
            ranks[x_i, y_i] = np.nan

    G: nx.Graph = nx.Graph()

    # Create weighted undirected graph
    pdistmatrix = pdist(ranks, metric=mk.distance)
    m, _ = ranks.shape

    elist = []
    for i, j in combinations(np.arange(m), r=2):
        idx = m * i + j - ((i + 2) * (i + 1)) // 2
        elist.append((i, j, 1 / (1 + pdistmatrix[idx])))

    G.add_weighted_edges_from(elist)

    # Compute the pagerank of each node
    pagerank = np.array(list(nx.pagerank(G).values()))

    return pagerank


def averagedistance(ranks, top_k=None) -> np.ndarray:
    """Computes the average distance of each rank to all other ranks.
    Lower average implies that a rank is more reliable.
    """
    if top_k is not None:
        ranks = ranks.astype(float)
        x, y = np.where((ranks > (top_k - 1)))
        for x_i, y_i in zip(x, y):
            ranks[x_i, y_i] = np.nan

    tol = 1e-6
    averagedist = squareform(pdist(ranks, metric=mk.distance)).mean(axis=1)
    return 1 / (tol + averagedist)



'''Tests for distributed_algorithm_instance.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=protected-access,duplicate-code,missing-function-docstring,redefined-outer-name
# pylint: disable=too-many-locals

from copy import deepcopy
from pathlib import Path
import pickle
import shutil
import socket
import time
from typing import Any, Dict, List, Mapping

import pandas as pd
import pytest

from ...algorithms.impl.neighbor_metadata import NeighborMetadata
from ...algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ...config_components.distributed_config import DistributedConfig
from ...conftest import Clearer, Waiter
from ...neighbor_manager.node_id import NodeID
from ...wrangler.base_port import BasePort
from ...wrangler.dataset import Dataset
from ...wrangler.logger import Logger

from ..distributed.auton_mean import AutonMean

from .distributed_algorithm_instance import DistributedAlgorithmInstance, Dropper
from .fake_distributed_algorithm import (
    FakeDistributedAlgorithm, FakeDistributedAlgorithmNeighbor, FitCountDistributedAlgorithm,
    FitCountDistributedAlgorithmNeighbor, FakeDistributedInstance)
from .synchronous import advance, fit_now, read_from_neighbors


logger = Logger(__file__, to_stdout=False).logger()


base_port = BasePort()


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("states")
    retval.mkdir(parents=True, exist_ok=True)
    return retval


LOOPBACK = '127.0.0.1'
SUNNY_DAY_SENDER = base_port.next()
SUNNY_DAY_RECIEVER = base_port.next()


def test_columns_sunny_day(wait_til_all_fit: Waiter,
                           assert_no_exceptions: Clearer) -> None:
    '''We get a neighbor report that has the same columns.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, SUNNY_DAY_RECIEVER)),
                    ('2', (LOOPBACK, SUNNY_DAY_SENDER))
                ]
            }
        },
        'my_id': 1,
    }

    dut = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=distributed_clause))

    fake_data = Dataset()
    fake_data.dataframe = pd.DataFrame({'a': [1], 'b': [2]})

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, SUNNY_DAY_SENDER))

    message = pickle.dumps(['a', 'b'])

    try:
        dut.start()
        dut.fit(fake_data)

        sock.sendto(message, (LOOPBACK, SUNNY_DAY_RECIEVER))

        # Wait for the message to arrive and get processed.
        wait_til_all_fit([dut])

        assert_no_exceptions([dut])

    finally:
        dut.stop()
        sock.close()


RAINY_DAY_SENDER = base_port.next()
RAINY_DAY_RECEIVER = base_port.next()


def test_columns_rainy_day(assert_no_exceptions: Clearer) -> None:
    '''We get a neighbor report that has mismatched columns.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, RAINY_DAY_RECEIVER)),
                    ('2', (LOOPBACK, RAINY_DAY_SENDER))
                ]
            }
        },
        'my_id': 1,
    }

    dut = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=distributed_clause),
        synchronous=True)

    fake_data = Dataset()
    fake_data.dataframe = pd.DataFrame({'a': [1], 'b': [2]})

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    try:
        sock.bind((LOOPBACK, RAINY_DAY_SENDER))
    except OSError as err:
        raise OSError(f'port: {RAINY_DAY_SENDER}') from err

    message = pickle.dumps(['a', 'b', 'c'])  # does not match

    try:
        dut.start()
        dut.fit(fake_data)

        sock.sendto(message, (LOOPBACK, RAINY_DAY_RECEIVER))

        # Wait for the message to arrive and get processed.
        advance([dut])

        got = dut.poll_exceptions()
        assert got is not None
        assert isinstance(got[0], NotImplementedError)
        assert 'mismatch' in str(got[1])

        assert_no_exceptions([dut])

    finally:
        dut.stop()
        sock.close()


REFIT_PORT = base_port.next()


def test_refit_with_no_message(wait_til_all_fit: Waiter,
                               assert_no_exceptions: Clearer) -> None:
    '''Distributed algs must refit if their state has changed, even if they receive no messages.

    Motivation: If a distributed algorithm has high self-regularization, its state may 'drift'
        when refit multiple times even with no new data or messages.
    If such an algorithm does not recieve messages for some reason (it has no neighbors or its
        neighbors are unresponsive), we still want to refit until its state is no longer changing.

    In FitCountDistributedAlgorithmInstance, the NeighborState has a property num_fits: int
    that changes on refit regardless of data or messages.

    should_send returns True if num_fits < 10.

    We create an instance with no neighbors and ensure that num_fits makes it to 10.
    '''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': []}
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, REFIT_PORT))
                ]
            }
        },
        'my_id': 1,
    }

    dut = FitCountDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=distributed_clause))

    try:
        dut.start()

        # The first fit doesn't trigger spontaneous refitting.
        dut.fit(None)

        # The second fit should trigger spontaneous refitting.
        dut.fit(None)
        # Convergence timer starts here since we have no neighbors

        assert wait_til_all_fit([dut], convergence_check=True), (
            'Timeout waiting for fit'
        )

        got = dut.my_state
        assert isinstance(got, FitCountDistributedAlgorithmNeighbor)
        assert got.fit_count == 3  # should_send = True if num_fits < 3

        assert_no_exceptions([dut])
    finally:
        dut.stop()


def test_synchronous(wait_til_all_fit: Waiter, assert_no_exceptions: Clearer) -> None:
    '''Test that setting 'synchronous=true' hyperparam causes nodes to not run fit loop.

    The FakeDistributedInstance will not set its state if it sees no data and no messages,
        but if fit on no data with a neighbor message will update its state to match the neighbor.

    We check that the neighbor message is not incorporated into the state
        until we tell the model to fit.
    '''

    sender_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, base_port.next())),
                    ('2', (LOOPBACK, base_port.next()))
                ]
            }
        },
        'my_id': 1,
    }
    reciever_clause = deepcopy(sender_clause)
    reciever_clause['my_id'] = 2

    sender = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=sender_clause),
        synchronous=True)

    receiver = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=reciever_clause),
        synchronous=True)

    input_dataset = Dataset(
        dataframe=pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6],
            'c': [7, 8, 9]
        })
    )

    try:
        sender.start()
        receiver.start()

        sender.fit(input_dataset)

        wait_til_all_fit([sender], convergence_check=False)

        # Make sure the receiver has received the message
        i: int = 0
        while not receiver._pending_message_from_neighbor:
            receiver.read_from_neighbors()
            assert i < 10, (
                'Timeout while waiting to recieve message. Probably a flake.'
            )
            time.sleep(0.5)
            i += 1

        # We have never told the receiver to fit.
        # Assuming fit_periodically is not running, it should not have fit or be fitting
        #   despite receiving a message.
        assert receiver.my_state is None
        assert not receiver._training

        receiver.fit(None)

        wait_til_all_fit([receiver], convergence_check=False)

        assert receiver.my_state
        assert receiver.my_state.columns == ['a', 'b', 'c']

        assert_no_exceptions([sender, receiver])
    finally:
        sender.stop()
        receiver.stop()


def test_synchronous_api(assert_no_exceptions: Clearer) -> None:
    '''Test synchronous API calls fit_now, and read_from_neighbors.'''

    sender_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, base_port.next())),
                    ('2', (LOOPBACK, base_port.next()))
                ]
            }
        },
        'my_id': 1,
    }
    reciever_clause = deepcopy(sender_clause)
    reciever_clause['my_id'] = 2

    sender = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=sender_clause),
        synchronous=True)

    receiver = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=reciever_clause),
        synchronous=True)

    input_dataset = Dataset(
        dataframe=pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6],
            'c': [7, 8, 9]
        })
    )

    try:
        sender.start()
        receiver.start()

        duts: List[DistributedAlgorithmInstance] = [sender, receiver]
        sender.fit(input_dataset)

        # Test read_from_neighbors.
        read_from_neighbors(duts, timeout_ns=1_000_000_000)

        # We have never told the receiver to fit.
        # Assuming fit_periodically is not running, it should not have fit or be fitting
        #   despite receiving a message.
        assert receiver.my_state is None
        assert not receiver._training

        # Test fit_now.
        fit_now(duts)

        assert receiver.my_state
        assert receiver.my_state.columns == ['a', 'b', 'c']

        assert sender.converged is False
        assert receiver.converged is False

        time.sleep(2.0)

        assert sender.converged is True
        assert receiver.converged is True

        assert_no_exceptions(duts)

    finally:
        sender.stop()
        receiver.stop()


def test_synchronous_advance(assert_no_exceptions: Clearer) -> None:
    '''Test synchronous API call advance.'''

    sender_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, base_port.next())),
                    ('2', (LOOPBACK, base_port.next()))
                ]
            }
        },
        'my_id': 1,
    }
    reciever_clause = deepcopy(sender_clause)
    reciever_clause['my_id'] = 2

    sender = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=sender_clause),
        synchronous=True)

    receiver = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=reciever_clause),
        synchronous=True)

    input_dataset = Dataset(
        dataframe=pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6],
            'c': [7, 8, 9]
        })
    )

    try:
        sender.start()
        receiver.start()

        duts: List[DistributedAlgorithmInstance] = [sender, receiver]
        # Test advance.
        advance(duts, timeout_ns=1_000_000_000)

        # Nobody had anythng to say.
        assert sender.my_state is None
        assert receiver.my_state is None

        sender.fit(input_dataset)

        assert sender.my_state is not None
        assert receiver.my_state is None

        advance(duts, min_time=1_000_000_000, timeout_ns=10_000_000_000)

        assert receiver.my_state == sender.my_state

        assert_no_exceptions(duts)

    finally:
        sender.stop()
        receiver.stop()


def _make_dropper(drop_rate: float, seed: int) -> Dropper:
    distributed_clause: Dict[str, Any] = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'stub_communicator',
        },
        'my_id': 1,
        'dropper': {
            'drop_rate': drop_rate,
            'seed': seed,
        }
    }

    config = DistributedConfig(clause=distributed_clause)
    assert config.dropper is not None
    return Dropper(config.dropper)


def test_dropper() -> None:
    '''Test that the dropper configuration can drop messages.'''

    dut1 = _make_dropper(1.0, 42)
    trial1 = sum(dut1.drop() for _ in range(100))
    assert trial1 == 99

    dut2 = _make_dropper(0.0, 42)
    trial2 = sum(dut2.drop() for _ in range(100))
    assert trial2 == 0

    dut3 = _make_dropper(0.5, 42)
    trial3 = sum(dut3.drop() for _ in range(100))
    assert trial3 == 47


@pytest.mark.skip(reason="We changed the location of emit_on_fit to periodic fit.")
def test_emit_on_fit(tmp_path: Path,
                     wait_til_all_fit: Waiter) -> None:
    '''Test that the emit_on_fit happens.'''
    distributed_clause: Dict[str, Any] = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': []}
            }
        },
        'communicator': {
            'name': 'stub_communicator',
        },
        'my_id': 1,
        'dropper': {
            'drop_rate': 0.0,
            'seed': 42,
            'output_dir': tmp_path,
        },
    }

    dut = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=distributed_clause),
        synchronous=True)

    input_dataset = Dataset(
        dataframe=pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6],
            'c': [7, 8, 9]
        })
    )

    try:
        dut.start()

        dut.fit(input_dataset)

        assert wait_til_all_fit([dut], convergence_check=False), 'Timeout waiting for fit'

        # Read in the file that was written by the dropper.
        got = list(tmp_path.glob('*.pkl'))
        assert len(got) == 1
        with open(got[0], 'rb') as f:
            data = pickle.load(f)
        assert data['now'] < time.monotonic_ns()
        assert data['fit_counter'] == 1
        assert data['fit_periodic_counter'] == 0
        assert data['predict_counter'] == 0
        assert data['dropper_tried_counter'] == 0
        assert data['dropper_dropped_counter'] == 0
        assert pickle.loads(data['my_state']) == ['a', 'b', 'c']
    finally:
        dut.stop()
        shutil.rmtree(tmp_path)


def test_multiple_emit_on_fit(tmp_path: Path,
                              wait_til_all_fit: Waiter,
                              assert_no_exceptions: Clearer) -> None:
    '''We generate a series of saved states.'''

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': []}
            }
        },
        'communicator': {
            'name': 'stub_communicator',
        },
        'my_id': 1,
        'dropper': {
            'drop_rate': 0.0,
            'seed': 42,
            'output_dir': tmp_path,
        },
    }

    dut = FitCountDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=distributed_clause))

    try:
        dut.start()

        # The first fit doesn't trigger spontaneous refitting.
        dut.fit(None)
        dut.fit(None)
        # Convergence timer starts here since we have no neighbors

        assert wait_til_all_fit([dut], convergence_check=True), (
            'Timeout waiting for convergence'
        )

        assert_no_exceptions([dut])

        got = list(tmp_path.glob('*.pkl'))
        assert len(got) >= 3
        # emit moved to fit_periodic, so we cannot predict the output
        # want = [
        #     {
        #         'my_id': 1,
        #         'fit_counter': 1,
        #         'fit_periodic_counter': 0,
        #         'dropper_tried_counter': 0,
        #         'dropper_dropped_counter': 0,
        #         'my_state': 1,
        #     },
        #     {
        #         'my_id': 1,
        #         'fit_counter': 2,
        #         'fit_periodic_counter': 0,
        #         'dropper_tried_counter': 0,
        #         'dropper_dropped_counter': 0,
        #         'my_state': 2,
        #     },
        #     {
        #         'my_id': 1,
        #         'fit_counter': 3,
        #         'fit_periodic_counter': 1,
        #         'dropper_tried_counter': 0,
        #         'dropper_dropped_counter': 0,
        #         'my_state': 3,
        #     },
        # ]
        # n = 0
        # print("DEBUG: got:", got)
        # for filename in sorted(got):
        #     print("DEBUG: n:", n)
        #     with open(filename, 'rb') as f:
        #         data = pickle.load(f)
        #     assert data['my_id'] == want[n]['my_id'], 'my_id mismatch'
        #     assert data['fit_counter'] == want[n]['fit_counter']
        #     assert data['fit_periodic_counter'] == want[n]['fit_periodic_counter']
        #     assert data['dropper_tried_counter'] == want[n]['dropper_tried_counter']
        #     assert data['dropper_dropped_counter'] == want[n]['dropper_dropped_counter']
        #     assert pickle.loads(data['my_state']) == want[n]['my_state']
        #     n += 1

    finally:
        dut.stop()
        shutil.rmtree(tmp_path)


DROP_ALL_NODE_1_PORT = base_port.next()
DROP_ALL_NODE_2_PORT = base_port.next()


@pytest.mark.skip(reason="We changed the location of emit_on_fit to periodic fit.")
def test_emit_on_fit_drop_all(tmp_path: Path,
                              wait_til_all_fit: Waiter,
                              assert_no_exceptions: Clearer) -> None:
    '''Drop every packet from a neighbor, but still emit on fit.'''
    node_1_train = Dataset(
        covariates=pd.DataFrame(
            {'a': [1.0, 2.0, 3.0], 'b': [2.0, 3.0, 4.0], 'c': [3.0, 4.0, 5.0]}))

    want_node_1 = pd.DataFrame({'a': [2.0], 'b': [3.0], 'c': [4.0]})

    node_2_train = Dataset(
        covariates=pd.DataFrame(
            {'a': [7.0, 8.0, 9.0], 'b': [8.0, 9.0, 10.0], 'c': [9.0, 10.0, 11.0]}))

    want_node_2 = pd.DataFrame({'a': [8.0], 'b': [9.0], 'c': [10.0]})

    placeholder_dataset = Dataset(
        covariates=pd.DataFrame(
            {'a': [0], 'b': [0], 'c': [0]}))

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, DROP_ALL_NODE_1_PORT)),
                    ('2', (LOOPBACK, DROP_ALL_NODE_2_PORT)),
                ]
            }
        },
        'my_id': 0,
        'dropper': {
            'drop_rate': 1.0,
            'output_dir': tmp_path,
        },

    }

    # TODO(Dan) Catch the case where there is only 1 class in the training data.
    distributed_node_1_clause = deepcopy(distributed_clause)
    distributed_node_1_clause['my_id'] = 1
    distributed_node_2_clause = deepcopy(distributed_clause)
    distributed_node_2_clause['my_id'] = 2

    alg = AutonMean()

    dut_node_1 = alg.instantiate(
        distributed=DistributedConfig(distributed_node_1_clause)
    )
    dut_node_2 = alg.instantiate(
        distributed=DistributedConfig(distributed_node_2_clause)
    )

    try:
        dut_node_1.start()
        dut_node_2.start()

        dut_node_1.fit(node_1_train)
        dut_node_2.fit(node_2_train)

        wait_til_all_fit([dut_node_1, dut_node_2], convergence_check=False)

        dut_node_1.fit(None)
        dut_node_2.fit(None)
        wait_til_all_fit([dut_node_1, dut_node_2], convergence_check=False)

        assert_no_exceptions([dut_node_1, dut_node_2])

        got_node_1 = dut_node_1.predict(placeholder_dataset)
        got_node_2 = dut_node_2.predict(placeholder_dataset)

        assert got_node_1 is not None
        assert got_node_2 is not None

        pd.testing.assert_frame_equal(got_node_1.predictions, want_node_1)
        pd.testing.assert_frame_equal(got_node_2.predictions, want_node_2)

        got_files = list(tmp_path.glob('*.pkl'))
        assert len(got_files) == 4
        n = 0
        want = [
            {'my_id': 1, 'dropper_dropped_counter': 0, 'dropper_tried_counter': 0},
            {'my_id': 1, 'dropper_dropped_counter': 1, 'dropper_tried_counter': 1},
            {'my_id': 2, 'dropper_dropped_counter': 0, 'dropper_tried_counter': 0},
            {'my_id': 2, 'dropper_dropped_counter': 1, 'dropper_tried_counter': 1},
        ]
        for filename in sorted(got_files):
            with open(filename, 'rb') as f:
                data = pickle.load(f)
            assert data['dropper_dropped_counter'] == want[n]['dropper_dropped_counter']
            assert data['dropper_tried_counter'] == want[n]['dropper_tried_counter']
            n += 1

    finally:
        dut_node_2.stop()
        dut_node_1.stop()
        shutil.rmtree(tmp_path)


def test_instantiate_with_state() -> None:
    '''Test that we can instantiate with a state.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': []}
            }
        },
        'communicator': {
            'name': 'stub_communicator',
        },
        'my_id': 1,
    }

    dataset = Dataset(
        dataframe=pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6],
            'c': [7, 8, 9]
        })
    )

    src = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=distributed_clause))

    src.fit(dataset)

    assert src.my_state is not None
    assert src.my_state.columns == ['a', 'b', 'c']

    dut = FakeDistributedInstance.instantiate_with_state(
        state=src.my_state.encode(),
        parent=FakeDistributedAlgorithm(),
        distributed=DistributedConfig(clause=distributed_clause))
    assert dut.my_state is not None, 'my_state is None'
    assert dut.my_state.columns == ['a', 'b', 'c']
    got = dut.predict(dataset)
    assert got is not None, 'predict returned None'
    pd.testing.assert_frame_equal(got.dataframe, dataset.dataframe)


DECIDER_0_PORT = base_port.next()
DECIDER_1_PORT = base_port.next()
DECIDER_2_PORT = base_port.next()


@pytest.mark.skip('This test is incomplete.')
def test_3_node_deciders(wait_til_all_fit: Waiter,
                         assert_no_exceptions: Clearer) -> None:
    '''3 sockets communicators on different nodes make decisions on whether to update each other.'''

    distributed: Dict[str, Any] = {
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('0', (LOOPBACK, DECIDER_0_PORT)),
                    ('1', (LOOPBACK, DECIDER_1_PORT)),
                    ('2', (LOOPBACK, DECIDER_2_PORT)),
                ]
            },
            'strategy': 'unicast',
        },
        'decider': {
            # It’s been too long since we last sent or received a model.
            'timesince_send': {
                'timeout': 2.0,  # seconds
            },
            'timesince_recv': {
                'timeout': 2.0,  # seconds
            },
            'distance': {
                # Model distance metric is high.
                'threshold': 5.0,  # Value is specific to algorithm.
            },
            'basic_risk': {
                # Risk has reduced sufficiently to make communication safe.
                'threshold': 0.1,  # Value is specific to communicator.
            },
        },
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '0': [1, 2],
                    '1': [0, 2],
                    '2': [0, 1],
                }
            }
        },
        'my_id': None,
    }

    duts: List[FakeDistributedInstance] = []

    alg = FakeDistributedAlgorithm()
    for i in range(3):
        config = deepcopy(distributed)
        config['my_id'] = i
        duts.append(alg.instantiate(distributed=DistributedConfig(config)))

    for i in range(3):
        duts[i].start()

    try:
        duts[0].fit(None)  # Maybe provide a special dataset?
        wait_til_all_fit(duts, convergence_check=False)
        # TODO(piggy): Check that only one of the neighbors is updated.

        duts[1].fit(None)  # Version 2 of the special dataset?
        wait_til_all_fit(duts, convergence_check=False)
        for i in [0, 2]:
            pass
            # Confirm that both nodes have updated.

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            if dut is not None:
                dut.stop()


@pytest.mark.parametrize(
    'threshold, enabled_distance, enabled_stub, result',
    [
        (0.5, False, False, False),
        (0.5, True, False, False),
        (-0.5, True, False, True),
        (0.5, False, True, True),
        (0.5, True, True, True),
    ]
)
def test_neighbor_decide(
        threshold: float, enabled_distance: bool, enabled_stub: bool, result: bool) -> None:

    config = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {'1': [2], '2': [1]}
            }
        },
        'communicator': {
            'name': 'stub_communicator',
        },
        'my_id': 1,
        'decider': {
            'distance': {
                'threshold': threshold,
                'enabled': enabled_distance,
            },
            'stub_decider': {
                'enabled': enabled_stub,
            },
        },
    }

    dut = FakeDistributedAlgorithm().instantiate(
        distributed=DistributedConfig(clause=config))
    dut._my_state_metadata = NeighborMetadata(my_state=FakeDistributedAlgorithmNeighbor())

    neighbor2_id = NodeID(2)
    neighbors: Mapping[NodeID, NeighborMetadataInterface] = {
        neighbor2_id: NeighborMetadata(my_state=FakeDistributedAlgorithmNeighbor()),
    }

    assert dut.decide(neighbor_id=neighbor2_id, neighbors=neighbors) == result



'''This file handles integration for all the steps in distributed algorithms.'''
import json
import logging
from queue import Queue
import threading
import time
from typing import Any, Callable, Dict, Iterable, Optional, Type

from flask import Flask, request

from ..algorithms.impl.distributed_algorithm_instance import DistributedAlgorithmInstance
from ..config_components.distributed_config import DistributedConfig
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import ExecutablePipeline
from ..metrics.impl.metric import Metric
from ..neighbor_manager.node_id import NodeID
from ..searcher.params import Override, Overrides, ParamRange, ParamRanges, Selector
from ..searcher.searcher import Searcher, SearcherImpl
from ..splitters.impl.splitter import Splitter
from ..templates.impl.pipeline_template import PipelineTemplate
from ..wrangler.dataset import Dataset, DatasetKeys
from ..wrangler.exception_thread import ExceptionThread

from .constants import Matcher, RangeMethod
from .wrangler_base import WranglerBase
from .wrangler_result import WranglerResult
from .logger import Logger, Level

log = Logger(__file__, level=logging.DEBUG, to_stdout=True).logger()

# pylint: disable=too-many-statements


class DistributedWrangler(WranglerBase):
    '''The wrangler is the central control object for distributed algorithms.'''
    _results: Queue[WranglerResult]
    _stop: bool = False
    _splitter: Splitter
    _dataloader_fit_thread: Optional[ExceptionThread] = None
    _async_fit_thread: Optional[ExceptionThread] = None
    _async_fit_lock: threading.Lock
    _async_fit_next_data: Optional[Dataset] = None

    def __init__(self, *args,
                 searcher: Optional[Type[Searcher]] = None,
                 my_id: Optional[int] = None,
                 **kwargs):
        log.info("Distributed wrangler starting.")
        self._results = Queue()
        self._async_fit_lock = threading.Lock()
        super().__init__(*args, **kwargs)

        # Inject the DistributedConfig as a hyperparam called 'distributed'
        #   for all distributed algorithms.
        distributed_overrides = Overrides()
        config = self._pd.get_conf('distributed')
        assert isinstance(config, DistributedConfig)
        if my_id is not None:
            config.my_id = NodeID(my_id)

        distributed_overrides.append(
            Override(
                selector=Selector({
                    Matcher.TAGS: {'distributed': 'true'}
                }),
                params=ParamRanges(distributed=ParamRange(
                    method=RangeMethod.FIXED,
                    prange=config
                )),
                no_show=True
            )
        )

        self._searcher = (searcher or SearcherImpl)(
            hyperparams=self._pd.hyperparams,
            inject=distributed_overrides)

        self._bound_pipelines = self._build_pipelines()
        executable_pipelines: Dict[Designator, ExecutablePipeline] = {
            pipeline.designator: self._instantiator_factory.instantiate(
                kind=self._executor.kind, pipeline=pipeline
            )
            for pipeline in self._bound_pipelines.values()
        }
        self._all_executable_pipelines = executable_pipelines
        self._current_executable_pipelines = executable_pipelines
        # We must init the data loader before we start the fitter thread.
        # The MemoryDataLoader can not search the stack for variables from a thread.
        self._find_data_loader()

    @property
    def all_distributed_instances(self) -> Iterable[DistributedAlgorithmInstance]:
        '''Fetch all the DistributedAlgorithmInstances.'''
        for pipeline in (self._current_executable_pipelines or {}).values():
            for instance in pipeline.all_instances:
                if isinstance(instance, DistributedAlgorithmInstance):
                    yield instance

    def wait_til_all_fit(self) -> None:
        '''Wait until all distributed instances have fit.'''
        all_trained = False
        while not all_trained:
            all_trained = True
            for instance in self.all_distributed_instances:
                # Note that trained takes the fit lock, so will
                # block until the fit is done.
                if not instance.trained:
                    all_trained = False
                    break

    def _dataloader_fit_loop(self, stop: Callable[[], bool]):
        '''This function is the data thread to process data.'''
        distributed = self._pd.get_conf('distributed')
        assert isinstance(distributed, DistributedConfig)
        while not stop():
            next_data = self._find_data_loader().poll(timeout=distributed.polling_interval)
            if next_data is not None:
                self.fit(next_data)  # sets self._current_trained_pipelines

    def _async_fit_loop(self, stop: Callable[[], bool]):
        '''This function is handles fit calls from the REST API asynchronously.'''
        distributed = self._pd.get_conf('distributed')
        assert isinstance(distributed, DistributedConfig)
        log.debug('async fit thread starting')
        while not stop():
            with self._async_fit_lock:
                next_data = self._async_fit_next_data
                self._async_fit_next_data = None
                log.log(Level.VERBOSE, "Checking async data")
            if next_data is None:
                time.sleep(distributed.polling_interval)
            else:
                log.debug('fitting')
                self.fit(next_data)  # sets self._current_trained_pipelines

    def lookup_templates(self) -> Dict[str, PipelineTemplate]:
        '''Look up the templates that match the problem definition.

        An example of how to specify the 'just', 'memory', and 'templated' pipeline loaders in
        the problem def:

        .. code-block:: Python

            {
                'distributed': {
                    'pipelines': {
                        'just': ['autonLogisticRegression', 'some other algorithm']
                        'memory': ['my_memory_pipeline']
                        'templated': [
                            {
                                'template': 'binary_tabular_classification',
                                'alg': 'AutonLogisticRegression'
                            },
                            {
                                'data_type': 'tabular',
                                'task': 'binary_classification',
                                'alg': 'some_other_alg'
                            }
                        ]
                    }
                }
            }
        '''
        distributed = self._pd.get_conf('distributed')
        assert isinstance(distributed, DistributedConfig)

        retval: Dict[str, PipelineTemplate] = {}

        for loader_name, arguments in distributed.pipelines.items():
            pipeline_loader = self._pipeline_loader_catalog.lookup_by_name(loader_name)
            for argument in arguments:
                if isinstance(argument, str):
                    argument = [argument]
                args = []
                kwargs = {}
                if isinstance(argument, list):
                    args = argument
                if isinstance(argument, dict):
                    kwargs = argument
                loaded_pipeline = pipeline_loader.load(*args, **kwargs)
                retval[loaded_pipeline.designator] = loaded_pipeline

        return retval

    def results(self) -> Queue[WranglerResult]:
        '''Return the results queue.'''
        return self._results

    def start(self) -> None:
        '''Start the wrangler threads.'''

        log.debug('DistributedWrangler starting wrangler threads: %x', id(self))
        self._stop = False
        assert self._all_executable_pipelines is not None
        for executable_pipeline in self._all_executable_pipelines.values():
            log.debug('DistributedWrangler starting pipeline: %s(%x)',
                      executable_pipeline, id(executable_pipeline))
            executable_pipeline.start()

        if self._dataloader_fit_thread is None:
            self._dataloader_fit_thread = ExceptionThread(target=self._dataloader_fit_loop, kwargs={
                'stop': lambda: self._stop
            })
        self._dataloader_fit_thread.start()

        if self._async_fit_thread is None:
            self._async_fit_thread = ExceptionThread(target=self._async_fit_loop, kwargs={
                'stop': lambda: self._stop
            })
        self._async_fit_thread.start()

    def stop(self) -> None:
        '''Stop the wrangler threads.'''
        log.debug('DistributedWrangler stopping wrangler threads: %x', id(self))
        self._stop = True
        if self._dataloader_fit_thread is not None:
            self._dataloader_fit_thread.join()
            self._dataloader_fit_thread = None
        if self._async_fit_thread is not None:
            self._async_fit_thread.join()
            self._async_fit_thread = None
        assert self._all_executable_pipelines is not None
        for executable_pipeline in self._all_executable_pipelines.values():
            log.debug('DistributedWrangler stopping pipeline: %s(%x)',
                      executable_pipeline, id(executable_pipeline))
            executable_pipeline.stop()

    def build_dataset_from_json(self, json_data: Dict) -> Dataset:
        '''Build a dataset from JSON data.'''
        return self._find_data_loader().build_dataset_from_json(json_data)

    def server(self, host='127.0.0.1', port=8080) -> Callable[[], None]:
        '''Start the REST node server.'''
        app = Flask(__name__)

        @app.get('/wrangler/v1.0/status')
        def status():
            return {'status': 'OK' if self.all_trained else 'UNTRAINED'}

        @app.route('/wrangler/v1.0/fit', methods=['GET', 'POST'])
        def fit():
            if request.method == 'GET':
                return status()
            if request.method == 'POST':
                try:
                    log.info('Calling fit REST API.')
                    dataset = self.build_dataset_from_json(request.json)
                    log.debug('Input Data: %s', dataset)
                    with self._async_fit_lock:
                        self._async_fit_next_data = dataset
                except BaseException as e:  # pylint: disable=broad-except
                    return Flask.make_response({'error': str(e)}, 500)
                return status()
            raise ValueError(f'Invalid request method {request.method}')

        @app.route('/wrangler/v1.0/predict', methods=['GET', 'POST'])
        def predict():
            if request.method == 'GET':
                return status()
            if request.method == 'POST':
                try:
                    log.info('Calling predict REST API.')
                    dataset = self.build_dataset_from_json(request.json)
                    log.debug('Input Data: %s', dataset)
                    result = self.predict(new_data=dataset)
                    log.debug('Predictions: %s', result)
                    retval = {}
                    for typed_des, res in result.items():
                        des = str(typed_des)
                        retval[des] = {}

                        pred = res.prediction
                        assert pred is not None

                        retval[des] = pred.to_prejson()

                    log.debug('Sending message with predictions: %s', retval)
                    log.debug("json.dumps(retval)=%s", json.dumps(retval))

                    return retval
                except BaseException as e:  # pylint: disable=broad-except
                    return {'error': str(e)}
            raise ValueError(f'Invalid request method {request.method}')

        @app.route('/wrangler/v1.0/predict_and_score', methods=['GET', 'POST'])
        def predict_and_score() -> Dict[str, Any]:
            if request.method == 'GET':
                return status()
            if request.method == 'POST':
                try:
                    log.info('Calling predict_and_score REST API.')
                    input_json = request.json
                    assert input_json is not None
                    dataset = self.build_dataset_from_json(input_json)
                    log.debug('Input Data: %s', dataset)

                    # Extract ground truth
                    ground_truth = dataset.output()
                    if DatasetKeys.GROUND_TRUTH.value in dataset:
                        ground_truth[DatasetKeys.GROUND_TRUTH.value] = dataset.pop(
                            DatasetKeys.GROUND_TRUTH.value
                        )

                    result = self.predict(new_data=dataset)
                    log.debug('Predictions: %s', result)

                    metrics: Dict[str, Metric] = {}
                    metric_names = input_json.get('metrics', [])
                    for m in metric_names:
                        metric: Metric = self._metric_catalog.lookup_by_name(m)
                        metrics[metric.name] = metric

                    # Evaluate a metric for each pipeline
                    retval: Dict[str, Dict[str, Any]] = {}
                    for typed_des, res in result.items():
                        pred = res.prediction

                        des = str(typed_des)
                        assert pred is not None, (
                            f'BUG: {res} prediction is None.'
                        )
                        retval[des] = pred.to_prejson()

                        if metrics:
                            retval[des]['metrics'] = {}
                            for metric_name, metric in metrics.items():
                                score = metric.calculate(pred=pred,
                                                         ground_truth=ground_truth)
                                retval[des]['metrics'][metric_name] = score
                    log.debug('Return json: %s', retval)
                    return retval
                except BaseException as e:  # pylint: disable=broad-except
                    return {'error': str(e)}
            raise ValueError(f'Invalid request method {request.method}')

        @app.route('/wrangler/v1.0/wait_til_all_fit', methods=['GET'])
        def wait_til_all_fit() -> Dict[str, Any]:
            if request.method == 'GET':
                self.wait_til_all_fit()
            return status()

        def retval():
            self.start()
            log.info('DistributedWrangler starting server on %s:%s', host, port)
            app.run(host=host, port=port)

        return retval



'''Metadata for a neighbor node.

This augments the NeighborState object.
'''

from copy import deepcopy
import time
from typing import Optional

from .neighbor_metadata_interface import NeighborMetadataInterface
from .neighbor_state_interface import NeighborStateInterface


class NeighborMetadata(NeighborMetadataInterface):
    '''Metadata for a neighbor node.'''

    def __init__(self, my_state: Optional[NeighborStateInterface] = None):
        '''Initialize the metadata.'''
        super().__init__()
        self._current_state = my_state

    _current_state: Optional[NeighborStateInterface] = None
    _last_state: Optional[NeighborStateInterface] = None
    _last_state_sent: Optional[NeighborStateInterface] = None

    _last_state_update_time: float = 0.0
    _last_state_sent_time: float = 0.0

    _cumulative_payload_sent: int = 0

    @property
    def current_state(self) -> Optional[NeighborStateInterface]:
        '''The current state of the neighbor.'''
        return self._current_state

    @current_state.setter
    def current_state(self, state: Optional[NeighborStateInterface]):
        '''Set the current state of the neighbor.'''
        self._last_state = self._current_state
        self._last_state_update_time = time.monotonic()
        self._current_state = state

    @property
    def last_state(self) -> Optional[NeighborStateInterface]:
        '''The last state of the neighbor.'''
        return self._last_state

    @property
    def last_state_update_time(self) -> float:
        '''The time the neighbor was last updated.'''
        return self._last_state_update_time

    @property
    def time_since_last_state_update(self) -> float:
        '''The time since the neighbor was last updated.'''
        return time.monotonic() - self._last_state_update_time

    @property
    def last_state_sent(self) -> Optional[NeighborStateInterface]:
        '''The last state that was sent to the neighbor.'''
        return self._last_state_sent

    @last_state_sent.setter
    def last_state_sent(self, state: Optional[NeighborStateInterface]):
        '''Set the last state that was sent to the neighbor.'''
        self._last_state_sent_time = time.monotonic()
        if state is not None:
            self._cumulative_payload_sent += state.payload_size
        self._last_state_sent = deepcopy(state)

    def sent_current_state(self):
        '''Mark the current state as sent.'''
        self.last_state_sent = self.current_state

    @property
    def last_state_sent_time(self) -> float:
        '''The time the last state was sent to any neighbor.'''
        return self._last_state_sent_time

    @property
    def time_since_last_state_sent(self) -> float:
        '''The time since the last state was sent to any neighbor.'''
        return time.monotonic() - self._last_state_sent_time

    @property
    def cumulative_payload_sent(self) -> int:
        '''The total payload sent to the neighbor.'''
        return self._cumulative_payload_sent



'''Object in charge of neighbor management for distributed algorithms.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from queue import Queue
from typing import Any, Dict, List, Optional, Set

import mlflow  # type: ignore

from ..discoverers.impl.discoverer import Discoverer, DiscovererStub
from .event import Event, Recv, NewNeighbor, NeighborTimeout
from .node_id import NodeID


class NeighborManager(metaclass=abc.ABCMeta):
    '''Object in charge of neighbor management for a distributed algorithm.

    It performs neighbor discovery, communication management, synchronization and logging.
    '''
    _default_timeout: Optional[float]
    _discoverer: Discoverer
    _queue: Queue
    name: str
    tags: Dict[str, Any]

    def __init__(self,
                 discoverer: Discoverer,
                 timeout: Optional[float] = None
                 ):
        self._discoverer = discoverer
        self._default_timeout = timeout
        self._queue = Queue()

    def stop(self) -> None:
        '''Stop all supporting threads.'''
        self._discoverer.stop()

    def start(self) -> None:
        '''Start all supporting threads.'''
        self._discoverer.start(queue=self._queue)

    @property
    def known_neighbors(self) -> Set[NodeID]:
        '''Return the set of known neighbors.'''
        return self._discoverer.communicator.known_neighbors

    def laplacian(self) -> List[List[int]]:
        '''Return the Laplacian of the network.'''
        return self._discoverer.laplacian()

    @abc.abstractmethod
    def synchronize(self) -> None:
        '''Synchronize with all peers.

        This function blocks until all neighbors have called synchonize.

        This is for starting tests at approximately the same time.
        '''

    def send_all(self, payload: bytes) -> None:
        '''Send a payload to all my neighbors, with logging.'''
        # I do not understand why, but this line appears to hang in
        # ngautonml/wrangler/distributed_wrangler_test.py::test_server_sunny_day.
        if mlflow.active_run() is not None:
            mlflow.log_metric(key='PayloadSize', value=len(payload))
        self._send_all(payload)

    @abc.abstractmethod
    def _send_all(self, payload: bytes) -> None:
        '''Send a payload to all my neighbors.

        The sending algorithm is responsible for creating the payload.
        '''

    def send(self, nodes: List[NodeID], payload: bytes) -> None:
        '''Send a paylod to a specific set of neighbors, with logging.'''
        mlflow.log_metric(key='PayloadSize', value=len(payload))
        self._send(nodes=nodes, payload=payload)

    @abc.abstractmethod
    def _send(self, nodes: List[NodeID], payload: bytes) -> None:
        '''Send a payload to a specific set of neighbors.'''

    def poll_for_events(self, timeout: Optional[float] = None) -> List[Event]:
        '''Check for any pending events, with logging.

        If timeout is not None, wait up to timeout for the next event to arrive.

        Events can be a payload arrives, a neighbor shows up or becomes unreachable,
        or other error conditions.
        '''

        retval = self._poll_for_events(timeout=timeout)
        for e in retval:
            if isinstance(e, Recv):
                mlflow.log_metric(key='RecvLen', value=len(e.payload))
                mlflow.log_metric(key='RecvNeighbor', value=e.neighbor)
            elif isinstance(e, NewNeighbor):
                mlflow.log_metric(key='NewNeighbor', value=e.neighbor)
            elif isinstance(e, NeighborTimeout):
                mlflow.log_metric(key='NeighborTimeout', value=e.neighbor)
            else:
                mlflow.log_metric(key='UnknownEvent', value=e.neighbor)
        return retval

    @abc.abstractmethod
    def _poll_for_events(self, timeout: Optional[float] = None) -> List[Event]:
        '''Check for any pending events.

        If timeout is not None, wait up to timeout for the next event to arrive.

        Events can be a payload arrives, a neighbor shows up or becomes unreachable,
        or other error conditions.
        '''


class NeighborManagerStub(NeighborManager):
    '''stub'''
    name = 'stub_neighbor_manager'
    tags: Dict[str, Any] = {}

    def __init__(self):
        super().__init__(discoverer=DiscovererStub())

    def synchronize(self) -> None:
        pass

    def _send_all(self, payload: bytes) -> None:
        pass

    def _send(self, nodes: List[NodeID], payload: bytes) -> None:
        pass

    def _poll_for_events(self, timeout: Optional[float] = None) -> List[Event]:
        return [Recv(neighbor=NodeID(1), payload=b'hello, world!')]



'''Output configuration'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path
from typing import Optional, Set, Dict, Any, List

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..config_components.impl.config_component import ConfigComponent, ConfigError
from ..wrangler.constants import FileType, Defaults, ProblemDefKeys
from ..executor.executor_kind import ExecutorKind


class OutputConfigError(ConfigError):
    '''Errors during building or validating OutputConfig.'''


class OutputConfig(ConfigComponent):
    '''Output configuration'''

    class Keys(AEnum):
        '''Top-level clauses in output component'''
        OUTPUT_PATH = 'path'
        FILE_TYPE = 'file_type'
        INSTATIATIONS = 'instantiations'

    def __init__(self, clause: Dict[str, Any], parents: Optional[List[str]] = None):
        parents = self._add_parent(parents, ProblemDefKeys.OUTPUT.value)
        super().__init__(clause=clause, parents=parents)

    def validate(self, **kwargs) -> None:
        if self.path is not None:
            if self.path.exists() and not self.path.is_dir():
                raise OutputConfigError(
                    f'Output path {self.path} is not a directory'
                )

            if not self.path.parent.exists():
                raise OutputConfigError(f'Parent of path {self.path} does not exist.')

            if not self.path.parent.is_dir():
                raise OutputConfigError(f'Parent of path {self.path} is not a directory.')

            # Confirm that we have the needed permissions to create this directory.
            self.path.mkdir(exist_ok=True)

        self.file_type  # pylint: disable=pointless-statement

        try:
            # will raise an error if any instantiations are not allowed
            _ = self.instantiations
        except NotImplementedError as err:
            raise OutputConfigError(err) from err

    @property
    def path(self) -> Optional[Path]:
        '''Path to the output folder specified in the problem def'''
        if self._exists(self.Keys.OUTPUT_PATH):
            return Path(self._get(self.Keys.OUTPUT_PATH)).expanduser().resolve()
        return None

    @property
    def instantiations(self) -> Set[ExecutorKind]:
        '''Kinds of instantiations specified in the problem'''
        retval = self._get_with_default(
            self.Keys.INSTATIATIONS, dflt=Defaults.INSTANTIANTIONS)
        return set(ExecutorKind(s) for s in retval)

    @property
    def file_type(self) -> FileType:
        '''Type of file for predictions to be saved in.'''
        retval = self._get_with_default(self.Keys.FILE_TYPE, dflt=Defaults.FILE_TYPE)
        try:
            return FileType[retval.upper()]
        except KeyError as exc:
            raise OutputConfigError(f'{retval} is not a valid file type') from exc



'''Test the TableCatalog'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..stub import StubTable
from .table_catalog import TableCatalogStub

# pylint: disable=missing-function-docstring,duplicate-code


def test_register() -> None:
    dut = TableCatalogStub()
    assert 'stub' == dut.register(obj=StubTable)


def test_register_name() -> None:
    dut = TableCatalogStub()
    stub_table = StubTable
    assert 'somenamedtable' == dut.register(
        obj=stub_table, name='SomeNamedTable')


def test_lookup_by_name() -> None:
    dut = TableCatalogStub()
    some_table = StubTable
    dut.register(obj=some_table, name='SomeNamedTable')
    assert dut.lookup_by_name('SomeNamedTable') == some_table


def test_lookup_one_by_tag() -> None:
    dut = TableCatalogStub()
    dut.register(obj=StubTable, name='NamedStubTable', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 1


def test_lookup_by_tag_fail() -> None:
    dut = TableCatalogStub()
    dut.register(obj=StubTable, name='NamedStubTable', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='different_value')) == 0


def test_lookup_multi_by_tag() -> None:
    dut = TableCatalogStub()
    stub_table = StubTable
    dut.register(obj=stub_table, name="Name1", tags={'k1': ['v1']})
    stub_table2 = StubTable
    dut.register(obj=stub_table2, name="Name2", tags={'k1': ['v1']})
    stub_table3 = StubTable
    dut.register(obj=stub_table3, name="Name3", tags={'k1': ['different_value']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 2



'''A "simple" executable parallel step'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, Iterable, Optional

from typing_extensions import Protocol

from ...algorithms.impl.algorithm_instance import AlgorithmInstance
from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import Designator
from ...templates.impl.parallel_step import ParallelStep
from ...templates.impl.pipeline_step import PipelineStep
from ...wrangler.dataset import Dataset
from ..cucumber import JarOfCucumbers
from .simple_executable_step import InstantiationError, SimpleExecutableStep

# pylint: disable=missing-function-docstring


class PipelineConstructor(Protocol):
    '''Match the signature of SimpleExecutablePipeline().'''
    # Should be __call__(self, pipeline: BoundPipeline) -> SimpleExecutablePipeline
    # but that would cause circular imports.
    def __call__(self, pipeline: BoundPipeline) -> Any:
        ...


class SimpleParallelExecutableStep(SimpleExecutableStep):
    '''Executable step for a bound parallel step

    We want to build and store SimpleExecutablePipelines, but that would lead
    to a circulate import. We use an Any placehold for SimpleExecutablePipeline
    in the _pipelines property below. We still need to build
    SimpleExecutablePipelines, so we pass the constructor to __init__ where
    we need to use it. The constructor takes a named parameter called pipeline.
    See PipelineConstructor for the standard pythonic solution to defining a
    callable with named arguments.
    '''
    # This is really a Dict[str, SimpleExecutablePipeline], but that would lead
    # to a circular import.
    _pipelines: Dict[str, Any]
    _trained: bool = False

    def __init__(self, step: PipelineStep, pipeline_constructor: PipelineConstructor):
        '''Build a SimpleParallelStep.

        Args:
          step: PipelineStep,
          pipeline_constructor needs to be SimpleExecutablePipeline.
        '''
        super().__init__(bound_step=step, model=None)
        if not isinstance(step, ParallelStep):
            raise InstantiationError(
                'SimpleParallelStep requires a ParallelStep for name '
                f'{self.pipeline_designator_component}')
        self._bound_step = step
        self._pipelines = {}
        for key in step.subpipeline_keys:
            self._pipelines[key] = pipeline_constructor(pipeline=step.subpipeline(key))

    def _init_model(self, **unused_kwargs):
        '''PipelineStep has no model.'''
        self._model = None
        self._model_instance = None

    def set_trained(self):
        super().set_trained()
        for pipeline in self._pipelines.values():
            pipeline.set_trained()

    def fit(self, *args, **kwargs):
        '''Call fit for all the pipelines.'''
        for pipe in self._pipelines.values():
            pipe.fit(*args, **kwargs)
        self._trained = True

    def predict(self, *args, dataset: Optional[Dataset] = None, **kwargs) -> Dataset:
        '''Apply all models to the dataset.

        Returns:
          Dataset where the keys are the names given to the constructor
          and the values are the Datasets from the respective pipelines.
          You generally want to use Connect to pull out the pieces needed
          by the next step.
        '''
        assert dataset is not None
        result = dataset.output()
        for k, pipe in self._pipelines.items():
            executor_result = pipe.predict(*args, dataset=dataset, **kwargs)
            result[k] = executor_result.prediction
        return result

    def cucumberize_all(self, pipeline_designator: Designator) -> JarOfCucumbers:
        models = JarOfCucumbers()
        for pipe in self._pipelines.values():
            # pipe is a SimpleExecutablePipeline
            models.update(pipe.cucumberize_all(parent_designator=pipeline_designator))
        return models

    @property
    def trained(self) -> bool:
        return self._trained

    @property
    def all_instances(self) -> Iterable[AlgorithmInstance]:
        '''Iterate through all algorithm instances.'''
        for pipe in self._pipelines.values():
            for instance in pipe.all_instances:
                assert isinstance(instance, AlgorithmInstance)
                yield instance



'''Catalog for objects that load data from one format to another.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, List, Optional, Type, Union

from ...catalog.catalog import CatalogLookupError
from ...catalog.memory_catalog import MemoryCatalog
from ...config_components.dataset_config import DatasetConfig

from .data_loader import DataLoader


class DataLoaderCatalog(MemoryCatalog[Type[DataLoader]], metaclass=abc.ABCMeta):
    '''Base class for DataLoader catalogs'''

    def register(self, obj: Type[DataLoader], name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        return super().register(obj, name, tags)

    def lookup_by_formats(self,
                          input_format: str,
                          loaded_format: str = 'pandas_dataframe') -> Dict[str, Type[DataLoader]]:
        '''Get a data loader based on input and loaded format.

        Args:

        :input_format: The format of the input data.
        :loaded_format: The format of the loaded data.
        '''
        return self.lookup_by_tag_and(input_format=input_format, loaded_format=loaded_format)

    def construct_instance(self, config: DatasetConfig) -> DataLoader:
        '''Lookup and construct the instance described by config.'''
        if config.config is not None:
            retval = self.lookup_by_name(name=config.config)
            assert isinstance(retval, type(DataLoader))
            return retval(config=config)
        loaders = self.lookup_by_formats(**config.dataloader_tags)
        if len(loaders) != 1:
            raise CatalogLookupError(
                f'Expected exactly one DataLoader for {config.dataloader_tags},'
                f' found {len(loaders)}'
            )
        return next(iter(loaders.values()))(config=config)


class DataLoaderCatalogStub(DataLoaderCatalog):
    '''stub'''



'''Test the TSAD aggregators'''

from typing import Optional

import pytest

from ..generator.bound_pipeline import BoundPipelineStub
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import PipelineResult, PipelineResults
from ..metrics.impl.metric import Metric, MetricStub
from ..ranker.ranker import Ranking, Rankings
from ..wrangler.dataset import Dataset

from .impl.aggregator_auto import AggregatorCatalogAuto

# pylint: disable-msg=R0801,duplicate-code


class MetricSumPred(Metric):
    '''Test metric returning the sum of the predictions (modulo 1000).'''
    _name = 'sum_pred_metric'
    _high = True

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        return sum(pred.values()) % 1000


class MetricMinVal(Metric):
    '''Test metric returning the minimum of the predictions.'''
    _name = 'min_val_metric'
    _high = True

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        return min(pred.values())


ranking_123 = ['pipeline_1', 'pipeline_2', 'pipeline_3']
ranking_132 = ['pipeline_1', 'pipeline_3', 'pipeline_2']
ranking_213 = ['pipeline_2', 'pipeline_1', 'pipeline_3']
ranking_231 = ['pipeline_2', 'pipeline_3', 'pipeline_1']
ranking_312 = ['pipeline_3', 'pipeline_1', 'pipeline_2']
ranking_321 = ['pipeline_3', 'pipeline_2', 'pipeline_1']


@pytest.mark.parametrize("aggregator_name,final_ranking", [
    ('tsad.rank_aggregation.borda', ranking_123),
    ('tsad.rank_aggregation.kemeny', ranking_123),
    ('tsad.rank_aggregation.partial_borda', ranking_123),
    ('tsad.rank_aggregation.trimmed_borda', ranking_123),
    ('tsad.rank_aggregation.trimmed_partial_borda', ranking_123),
])
def test_tsad_simple_ranking(aggregator_name, final_ranking) -> None:
    '''Tests the TSAD rank aggregation methods.'''
    catalog = AggregatorCatalogAuto()
    aggregator = catalog.lookup_by_name(aggregator_name)

    results_1 = Dataset({
        '1': 1,
        '2': 2,
        '3': 3,
    })
    results_2 = Dataset({
        '1': 6,
        '2': 4,
        '3': 2,
    })
    results_3 = Dataset({
        '1': 7,
        '2': 5,
        '3': 11,
    })

    pipeline_results = PipelineResults({
        Designator('pipeline_1'): PipelineResult(prediction=results_1,
                                                 bound_pipeline=BoundPipelineStub('pipeline_1')),
        Designator('pipeline_2'): PipelineResult(prediction=results_2,
                                                 bound_pipeline=BoundPipelineStub('pipeline_2')),
        Designator('pipeline_3'): PipelineResult(prediction=results_3,
                                                 bound_pipeline=BoundPipelineStub('pipeline_3')),
    })

    rank_1 = Ranking(metric=MetricSumPred(), results=pipeline_results)
    rank_2 = Ranking(metric=MetricStub(), results=pipeline_results)
    rank_3 = Ranking(metric=MetricMinVal(), results=pipeline_results)

    rankings = Rankings({'rank_1': rank_1, 'rank_2': rank_2, 'rank_3': rank_3})

    aggregate_ranking = aggregator.aggregate(rankings=rankings, all_scores=True)

    assert [s.score for s in aggregate_ranking.scores(all_scores=True)] == [2, 1, 0]
    assert [s.pipeline_des for s in aggregate_ranking.scores(all_scores=True)] == final_ranking



'''Ranks pipelines by using metrics to evauluate results'''
import abc
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from ..generator.bound_pipeline import BoundPipelineStub
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import PipelineResult, PipelineResults
from ..metrics.impl.metric import Metric, MetricInvalidDatasetError, MetricStub
from ..splitters.impl.splitter import SplitDataset
from ..wrangler.constants import OutputColName
from ..wrangler.dataset import Dataset, DatasetKeys

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class RankerError(Exception):
    '''Base class for all errors related to Ranker'''


class ScoredResult():
    '''The results of running a single executable pipeline, with a score from a single metric'''
    _metric: Metric
    _result: PipelineResult
    _score: Optional[float]
    _unscorable_reason: Optional[str]

    def __init__(self,
                 metric: Metric,
                 result: PipelineResult,
                 score: Optional[float] = None,
                 unscorable_reason: Optional[str] = None):
        self._metric = metric
        self._result = result
        self._score = score
        self._unscorable_reason = unscorable_reason

    @property
    def pipeline_des(self) -> Designator:
        '''Designator for the pipeline that produced this ScoredResult'''
        return self._result.bound_pipeline.designator

    @property
    def family_des(self) -> Designator:
        '''Family Designator of the pipeline that produced this ScoredResult'''
        return self._result.bound_pipeline.family_designator

    @property
    def metric(self) -> Metric:
        '''Metric used to evaulate this score'''
        return self._metric

    @property
    def result(self) -> PipelineResult:
        '''PipelineResult used to evaulate this score'''
        return self._result

    @property
    def score(self) -> Optional[Union[float, str]]:
        '''A float representing this result's score,
        or a str explaining why it cannot be scored,
        or None if neither was provided.'''
        if self._score is not None:
            return self._score

        return self._unscorable_reason

    def __lt__(self, other: Any) -> bool:
        # a < b means "a is worse performance than b"
        if not isinstance(other, ScoredResult):
            raise NotImplementedError(
                '__lt__ for ScoredResult requires right-hand side to be a ScoredResult.\n'
                f'found: {other} of type {type(other)}.')

        if other._metric.name != self._metric.name:
            raise RankerError(
                'Cannot compare ScoredResults with differing metrics.\n'
                f'Left-hand side: {self._metric.name}'
                f'Right-hand side: {other._metric.name}')

        if other._score is None and self._score is None:
            # sort alphabetically by unscorable reason
            # (later in the alphabet = worse)
            self_reason = self._unscorable_reason or ""
            other_reason = other._unscorable_reason or ""
            return self_reason > other_reason

        if other._score is None:
            # scorable (self) > unscorable (other)
            return False

        if self._score is None:
            # unscorable (self) < scorable (other)
            return True

        if self._metric.high:
            return self._score < other._score

        return self._score > other._score

    def __str__(self) -> str:

        if self._score is None:
            if self._unscorable_reason is None:
                score_str = 'unscorable'
            else:
                score_str = f'unscorable ({self._unscorable_reason})'
        else:
            score_str = f'{self._score:.4f}'

        return (f'\tpipeline: {self._result.bound_pipeline!s}\n'
                f'\tscore: {score_str}')

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ScoredResult):
            return False
        return self._metric.name == other._metric.name and self._score == other._score


class Ranking:
    '''Holds the result of applying a single metric to multiple ExecutorResult.'''
    _metric: Metric
    _results: PipelineResults
    _scores: List[ScoredResult]
    _matriarch_scores: List[ScoredResult]
    _ground_truth: Optional[Dataset] = None

    def __init__(self,
                 metric: Metric,
                 results: PipelineResults,
                 ground_truth: Optional[Dataset] = None):
        self._metric = metric
        self._results = results
        self._scores = []  # List of ScoredResults, ordered from best to worst
        self._matriarch_scores = []
        self._ground_truth = ground_truth
        self.score()

    def __str__(self):
        retval = f'--{self._metric}--\n'
        for i, score in enumerate(self.scores()):
            retval = f'{retval}{i+1}:\n{score}\n'
        return retval

    @property
    def str_all_scores(self) -> str:
        '''Create a string representation of all scores.'''
        retval = f'--{self._metric}--\n'
        for i, score in enumerate(self.scores(all_scores=True)):
            retval = f'{retval}{i+1}:\n{score}\n'
        return retval

    def score(self) -> None:
        '''Calculate the scores of all results.'''
        self._scores = []
        family_designators = set()
        for res in self._results.values():
            family_designators.add(res.family_designator)
            if DatasetKeys.ERROR.value in (res.prediction or {}):
                assert res.prediction is not None, (
                    'BUG: prediction is None but error is somehow present.')
                self._scores.append(ScoredResult(
                    metric=self._metric,
                    result=res,
                    unscorable_reason=str(res.prediction[DatasetKeys.ERROR.value])
                ))
                continue
            try:
                if res.prediction is None:
                    raise MetricInvalidDatasetError('Prediction is None')
                score = self._metric.calculate(
                    pred=res.prediction,
                    ground_truth=self._ground_truth)
                scored_result = ScoredResult(
                    metric=self._metric,
                    result=res,
                    score=score)
            except MetricInvalidDatasetError as err:
                scored_result = ScoredResult(
                    metric=self._metric,
                    result=res,
                    unscorable_reason=str(err))
            self._scores.append(scored_result)

        # We want the best scores first
        # __lt__ on a ScoredResult is defined so that higher=better
        # Thus we always reverse (sort from high to low)
        self._scores.sort(reverse=True)

        # Pick out the highest scoring pipelines in each family,
        # the "matriarchs".
        for a_score in self._scores:
            if a_score.family_des in family_designators:
                self._matriarch_scores.append(a_score)
                family_designators.remove(a_score.family_des)

    @property
    def metric(self) -> Metric:
        '''The metric used to evaulate this Ranking'''
        return self._metric

    def scores(self, all_scores: bool = False) -> List[ScoredResult]:
        '''All ScoredResults, sorted from best to worst.'''
        if len(self._scores) != len(self._results):
            self.score()
        if all_scores:
            return self._scores
        return self._matriarch_scores

    def scores_as_dict(self, all_scores: bool = False) -> Dict[Designator, ScoredResult]:
        '''All ScoredResults, in a dict keyed by their pipeline's Designator.'''
        return {s.pipeline_des: s for s in self.scores(all_scores)}

    def as_dataframe(self, all_scores: bool = False) -> pd.DataFrame:
        '''Return ranking as a dataframe with 2 columns, pipeline designator and score.'''
        scores = self.scores(all_scores)
        retval = pd.DataFrame()
        retval[OutputColName.DESIGNATOR.value] = [str(s.pipeline_des) for s in scores]
        retval[self.metric.name] = [s.score for s in scores]
        return retval

    def best(self, num: int, all_scores: bool = False) -> List[ScoredResult]:
        '''Return best [num] ScoredResults, in order.'''
        if num < 0 or num > len(self.scores(all_scores=all_scores)):
            raise IndexError(
                f'Cannot get best {num} scores.  '
                f'Choose a number between 0 and {len(self.scores(all_scores=all_scores))}.')
        return self.scores(all_scores=all_scores)[:num]


class RankingStub(Ranking):
    '''stub'''

    def best(self, num: int, all_scores: bool = False) -> List[ScoredResult]:
        return [ScoredResult(
            metric=self._metric,
            result=PipelineResult(
                prediction=Dataset(),
                bound_pipeline=BoundPipelineStub(name='stub'),
                split_dataset=SplitDataset()),
            score=0.0)]


class Rankings(Dict[str, Ranking]):
    '''Dictionary of Rankings keyed by metric name.'''

    def __str__(self):
        summary = ['The rankings are:']
        for metric in sorted(self.keys()):
            summary.append(f'{self[metric]}')
        return '\n'.join(summary)

    def as_dataframe(self, order_metric: Optional[str] = None) -> pd.DataFrame:
        '''Return Rankings as a dataframe with columns for pipe designator and each metric.

        order_metric determines the order of pipelines in the output.
        If not provided, the first metric alphabetically will be used.
        '''
        metrics = sorted(self.keys())
        if order_metric is None:
            order_metric = metrics[0]
        retval = self[order_metric].as_dataframe().set_index(OutputColName.DESIGNATOR.value)
        for m in sorted(set(self.keys()) - {order_metric}):
            m_df = pd.DataFrame(self[m].as_dataframe())
            # Join additional metrics using designator as index, so that
            #   scores correspond to pipelines correctly.
            retval = retval.join(m_df.set_index(OutputColName.DESIGNATOR.value))
        # Change designator from index back to normal col
        retval.reset_index(inplace=True)
        return retval


class Ranker(metaclass=abc.ABCMeta):
    '''Base class for pipeline rankers'''

    def rank(self,
             results: PipelineResults,
             metrics: Dict[str, Metric],
             ground_truth: Optional[Dataset] = None) -> Rankings:
        '''Rank the results of a run of the executor using one or multiple metrics'''
        return Rankings({k: Ranking(v, results, ground_truth) for k, v in metrics.items()})


class RankerStub(Ranker):
    '''stub'''

    def rank(self,
             results: PipelineResults,
             metrics: Dict[str, Metric],
             ground_truth: Optional[Dataset] = None) -> Rankings:
        return Rankings({'stub metric': Ranking(
            metric=MetricStub(),
            results=PipelineResults({
                Designator('stub_designator'): PipelineResult(
                    prediction=Dataset(),
                    bound_pipeline=BoundPipelineStub(name='stub'),
                    split_dataset=SplitDataset()
                )
            })
        )})



'''Tests for the Metric Base class implementations'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional

import pytest

from ...catalog.memory_catalog import CatalogNameError
from ...wrangler.dataset import Dataset
from .metric import Metric

# pylint: disable=missing-function-docstring,duplicate-code
# pylint: disable=missing-class-docstring
# pylint: disable=too-few-public-methods


class ExampleMetric(Metric):
    _name = 'TestMetric'
    _high = True

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        return 0.0


def test_missing_name():

    class BadTestMetric(Metric):
        def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
            return 0.0

    with pytest.raises(CatalogNameError):
        BadTestMetric()


def test_name():
    test_metric = ExampleMetric()
    assert test_metric.name == 'testmetric'



'''Simple Domain Confusion'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Optional

import numpy as np

from ...catalog.catalog import upcast
from ...wrangler.dataset import Dataset, DatasetKeys
from ...wrangler.logger import Logger

from ..impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..impl.fittable_algorithm_instance import FittableAlgorithmInstance

logger = Logger(__file__).logger()


class AutonMomentMatcherInstance(FittableAlgorithmInstance):
    '''Instance of Simple Domain Confusion.

    This is a Domain Adaptation algorithm, that allows the use of a pretrained
    model adapted for a new datasaet.

    AutonSimpleDomainConfusion uses the mean and variance of a new dataset to
    adapt a model trained on a different dataset. The model is adapted by
    adjusting the the mean and variance of the new dataset to match the mean
    and variance of the original dataset.
    '''
    _A: np.ndarray
    _b: np.ndarray
    _hyperparam_overrides: Dict[str, Any]

    def __init__(self, parent: Algorithm, **hyperparams):
        self._hyperparam_overrides = {}

        mean = hyperparams.pop('mean', None)
        covariance = hyperparams.pop('covariance', None)
        if mean is not None:
            self._hyperparam_overrides['mean'] = mean
        if covariance is not None:
            self._hyperparam_overrides['covariance'] = covariance
        super().__init__(parent=parent, **hyperparams)

    def hyperparams(self, **hyperparams) -> Dict[str, Any]:
        '''Hyperparam override order:

        1. hyperparams from the Algorithm are overridden by
        2. hyperparams from __init__ are overridden by
        3. hyperparams from the dataset
        '''
        errors: List[str] = []
        overrides = self._hyperparam_overrides.copy()
        overrides.update(hyperparams)
        retval = self.algorithm.hyperparams(**overrides)
        if retval['mean'] is None:
            errors.append(
                f'mean is a required hyperparam for {self.algorithm.name}')
        if retval['covariance'] is None:
            errors.append(
                f'covariance is a required hyperparam for {self.algorithm.name}')
        if errors:
            raise ValueError('\n'.join(errors))
        return retval

    def fit(self, dataset: Optional[Dataset]) -> None:
        if dataset is None:
            return

        self._trained = True

        if DatasetKeys.HYPERPARAMS.value in dataset:
            hyperparams = self.hyperparams(**dataset[DatasetKeys.HYPERPARAMS.value])
        else:
            hyperparams = self.hyperparams()

        mu_source = hyperparams['mean']
        cov_source = hyperparams['covariance']

        numeric = dataset.dataframe.select_dtypes(include=['number'])

        mu_target = numeric.mean(axis=0)
        cov_target = np.cov(numeric.T)

        epsilon = 1e-9
        vals, vecs = np.linalg.eigh(cov_target)
        vals[vals < 0] = 0
        root_target = vecs @ np.diag(np.sqrt(vals))
        vals, vecs = np.linalg.eigh(cov_source)
        vals[vals < 0] = 0
        root_source = vecs @ np.diag(np.sqrt(vals))
        n, _ = root_target.T.shape
        # pylint: disable-next=invalid-name
        self._A = np.linalg.solve(root_target.T + epsilon * np.eye(n), root_source.T).T
        self._b = (mu_source - self._A @ mu_target).to_numpy()

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        new_dataset = dataset.output()
        columns = dataset.dataframe.columns
        numeric = dataset.dataframe.select_dtypes(include=['number'])
        nonnumeric = dataset.dataframe.select_dtypes(exclude=['number'])
        new_dataset.dataframe = numeric @ self._A.T + self._b[np.newaxis, :]
        new_dataset.dataframe = new_dataset.dataframe.join(nonnumeric)
        new_dataset.dataframe = new_dataset.dataframe.rename(
            columns=dict(zip(new_dataset.dataframe.columns, columns)))
        return new_dataset


class AutonMomentMatcher(Algorithm):
    '''Class for Auton Lab's implementation of Simple Domain Confusion.'''
    _name = "auton_simple_domain_confusion"
    _tags = {
        'source': ['auton_lab'],
        'preprocessor': ['true'],
    }
    _instance_constructor = AutonMomentMatcherInstance
    _default_hyperparams = {
        'mean': None,
        'covariance': None,
    }

    def instantiate(self, **hyperparams) -> 'AutonMomentMatcherInstance':
        return super().instantiate(**hyperparams)


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = AutonMomentMatcher(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Tests for distance.py'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


from typing import Mapping

from ..algorithms.impl.neighbor_metadata import NeighborMetadata
from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..algorithms.impl.toy_neighbor_state import ToyNeighborState
from ..config_components.distributed_config import DistributedConfig
from ..deciders.impl.decider import Decider
from ..neighbor_manager.node_id import NodeID

from .impl.decider_auto import DeciderCatalogAuto

# pylint: disable=duplicate-code


def test_sunny_day() -> None:
    '''Test the max_distance decider.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'max_distance': {
                    'num_neighbors': 1,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('max_distance')(
        config=distributed.decider.max_distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=my_state)
    near_id = NodeID(1)
    below_id = NodeID(2)
    distant_id = NodeID(3)
    neighbors = {
        near_id: NeighborMetadata(my_state=ToyNeighborState(center=0.1)),
        below_id: NeighborMetadata(my_state=ToyNeighborState(center=0.3)),
        distant_id: NeighborMetadata(my_state=ToyNeighborState(center=1.0)),
    }

    assert decider.decide(my_meta=my_meta, neighbor_id=near_id, neighbors=neighbors) is False
    assert decider.decide(my_meta=my_meta, neighbor_id=below_id, neighbors=neighbors) is False
    assert decider.decide(my_meta=my_meta, neighbor_id=distant_id, neighbors=neighbors) is True


def test_max_distance_no_state() -> None:
    '''Test the decider_stub in the DeciderCatalog class.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'max_distance': {
                    'num_neighbors': 1,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('max_distance')(
        config=distributed.decider.max_distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=my_state)
    near_id = NodeID(1)
    below_id = NodeID(2)
    distant_id = NodeID(3)
    neighbors = {
        near_id: NeighborMetadata(my_state=ToyNeighborState(center=0.1)),
        below_id: NeighborMetadata(my_state=None),
        distant_id: NeighborMetadata(my_state=None),
    }

    assert decider.decide(my_meta=my_meta, neighbor_id=near_id, neighbors=neighbors) is False
    assert decider.decide(my_meta=my_meta, neighbor_id=below_id, neighbors=neighbors) is True
    assert decider.decide(my_meta=my_meta, neighbor_id=distant_id, neighbors=neighbors) is True


def test_max_distance_no_neighbors() -> None:
    '''Test the decider_stub in the DeciderCatalog class.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'max_distance': {
                    'num_neighbors': 1,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('max_distance')(
        config=distributed.decider.max_distance)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=my_state)
    neighbor_id = NodeID(1)
    neighbors: Mapping[NodeID, NeighborMetadataInterface] = {}

    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is False
